# -*- coding:utf-8 -*-

"""this module is implement to deal the tracker all data
"""

import copy

import logging
import threading
from tingyun.logistics.workshop.quantile import QuantileP2
from tingyun.logistics.workshop.packets import TimePackets, ApdexPackets, SlowSqlPackets
from tingyun.packages import six


console = logging.getLogger(__name__)


class Packager(object):
    """
    """
    def __init__(self):
        self.__max_error_count = 20
        self.__settings = None

        self.__time_packets = {}  # store time packets data key is name + scope
        self.__apdex_packets = {}
        self.__action_packets = {}
        self.__general_packets = {}
        # it's maybe include {name: {trackerType, count, $ERROR_ITEM}}
        # $ERROR_ITEM  detail, check the documentation
        self.__traced_errors = {}

        # it's maybe include {name: {count, $ERROR_ITEM}}
        # $ERROR_ITEM  detail, check the documentation
        self.__traced_external_errors = {}

        # format: {$"metric_name": [$slow_action_data, $duration]}
        self.__slow_action = {}
        self.__slow_sql_packets = {}

        # quantile
        self.__quantile_data = None
        self.__quantile = {}

        self._packets_lock = threading.Lock()

    @property
    def settings(self):
        """
        :return:
        """
        return self.__settings

    def create_data_zone(self):
        """create empty data engine to contain the unmerged tracker data
        :return:
        """
        zone = Packager()
        zone.__settings = self.__settings

        return zone

    def reset_packets(self, application_settings):
        """
        :param application_settings: application settings from server and setting file
        :return:
        """
        self.__settings = application_settings

        self.__time_packets = {}
        self.__apdex_packets = {}
        self.__action_packets = {}
        self.__general_packets = {}
        self.__traced_errors = {}
        self.__traced_external_errors = {}
        self.__slow_action = {}
        self.__slow_sql_packets = {}

    def record_tracker(self, tracker):
        """
        :param tracker: tracker node instance
        :return:
        """
        if not self.__settings:
            console.error("The application settings is not merge into data engine.")
            return

        node_limit = self.__settings.action_tracer_nodes
        threshold = self.__settings.action_tracer.action_threshold

        self.record_time_metrics(tracker.time_metrics())  # deal for component, include framework/db/other
        self.record_quantile(*tracker.quantile())
        self.record_action_metrics(tracker.action_metrics())  # deal for user action
        self.record_apdex_metrics(tracker.apdex_metrics())  # for recording the apdex
        self.record_traced_errors(tracker.traced_error())  # for error trace detail
        self.record_slow_action(tracker.slow_action_trace(node_limit, threshold))
        self.record_slow_sql(tracker.slow_sql_nodes())
        self.record_traced_external_trace(tracker.traced_external_error())

    def record_quantile(self, action, duration):
        """
        :return:
        """
        if 0 == len(self.__settings.quantile):
            return

        self.__quantile_data = (action, duration)

    def record_slow_sql(self, nodes):
        """
        :param nodes: the slow sql node
        :return:
        """
        if not self.__settings.action_tracer.slow_sql:
            return

        for node in nodes:
            if len(self.__slow_sql_packets) > self.__settings.slow_sql_count:
                console.debug("Slow sql count is more than max count %s", self.__settings.slow_sql_count)
                continue

            key = node.request_uri + node.sql
            packets = self.__slow_sql_packets.get(key)
            if not packets:
                packets = SlowSqlPackets()

            packets.merge_slow_sql_node(node)
            self.__slow_sql_packets[key] = packets

    def record_slow_action(self, slow_action):
        """
        :param slow_action:
        :return:
        """
        if not self.__settings.action_tracer.enabled:
            return

        # 不能在这里通过判断阈值是否到达临界去除法慢过程，因为可能因为跨应用而强制触发
        # 如果节点返回`无效数据`则直接丢弃
        if not slow_action:
            return

        top_n = self.__settings.action_tracer.top_n
        metric_name = slow_action[2]

        if metric_name not in self.__slow_action:
            self.__slow_action[metric_name] = [[slow_action, slow_action[1]]]
        else:
            # every request url/metric can save top_n action trace.
            if len(self.__slow_action[metric_name]) <= top_n:
                self.__slow_action[metric_name].append([slow_action, slow_action[1]])

    def record_traced_errors(self, traced_errors):
        """
        :return:
        """
        if not self.__settings.error_collector.enabled:
            return

        for error in traced_errors:
            if len(self.__traced_errors) > self.__max_error_count:
                console.debug("Error trace is reached maximum limitation.")
                break

            if error.error_filter_key in self.__traced_errors:
                self.__traced_errors[error.error_filter_key]["count"] += 1
                self.__traced_errors[error.error_filter_key]["item"][-3] += 1
            else:
                self.__traced_errors[error.error_filter_key] = {"count": 1,
                                                                "item": error.trace_data,
                                                                "tracker_type": error.tracker_type}

    def record_traced_external_trace(self, traced_errors):
        """
        :param traced_errors:
        :return:
        """
        if not self.__settings.error_collector.enabled:
            return

        for error in traced_errors:
            if len(self.__traced_external_errors) > self.__max_error_count:
                console.debug("External error trace is reached maximum limitation.")
                break

            if error.error_filter_key in self.__traced_external_errors:
                self.__traced_external_errors[error.error_filter_key]["count"] += 1
                self.__traced_external_errors[error.error_filter_key]["item"][-3] += 1

            else:
                self.__traced_external_errors[error.error_filter_key] = {"count": 1, "status_code": error.status_code,
                                                                         "item": error.trace_data,
                                                                         "tracker_type": error.tracker_type}

    def record_general_metric(self, metric):
        """
        :param metric:
        :return:
        """
        key = (metric.name.split("/", 1)[1], '')

        packets = self.__general_packets.get(key)
        if packets is None:
            packets = TimePackets()

        packets.merge_time_metric(metric)
        self.__general_packets[key] = packets

        return key

    def record_time_metric(self, metric):
        """
        :param metric:
        :return:
        """
        # filter the general data from the metric, the metric node should be distinguish general and basic metric
        if metric.name.startswith("GENERAL"):
            self.record_general_metric(metric)
            return

        key = (metric.name, metric.scope or '')  # metric key for protocol
        packets = self.__time_packets.get(key)
        if packets is None:
            packets = TimePackets()

        packets.merge_time_metric(metric)
        self.__time_packets[key] = packets

        return key

    def record_time_metrics(self, metrics):
        """
        :param metrics:
        :return:
        """
        for metric in metrics:
            self.record_time_metric(metric)

    def record_apdex_metric(self, metric):
        """
        :param metric:
        :return:
        """
        key = (metric.name, "")
        packets = self.__apdex_packets.get(key)

        if packets is None:
            packets = ApdexPackets(apdex_t=metric.apdex_t)

        packets.merge_apdex_metric(metric)
        self.__apdex_packets[key] = packets
        return key

    def record_apdex_metrics(self, metrics):
        """
        :param metrics:
        :return:
        """
        for metric in metrics:
            self.record_apdex_metric(metric)

    def record_action_metric(self, metric):
        """
        :param metric:
        :return:
        """
        key = (metric.name, metric.scope or '')  # metric key for protocol
        packets = self.__action_packets.get(key)
        if packets is None:
            packets = TimePackets()

        packets.merge_time_metric(metric)
        self.__action_packets[key] = packets

        return key

    def record_action_metrics(self, metrics):
        """
        :param metrics:
        :return:
        """
        for metric in metrics:
            self.record_action_metric(metric)

    def rollback(self, stat, merge_performance=True):
        """rollback the performance data when upload the data failed. except the traced error count.
        :param stat:
        :param merge_performance:
        :return:
        """
        if not merge_performance:
            return

        console.warning("Agent will rollback the data which is captured at last time. That indicates your network is"
                        " broken.")

        for key, value in six.iteritems(stat.__time_packets):
            packets = self.__time_packets.get(key)
            if not packets:
                self.__time_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(stat.__apdex_packets):
            packets = self.__apdex_packets.get(key)
            if not packets:
                self.__apdex_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(stat.__action_packets):
            packets = self.__action_packets.get(key)
            if not packets:
                self.__action_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(stat.__general_packets):
            packets = self.__general_packets.get(key)
            if not packets:
                self.__general_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(stat.__traced_errors):
            packets = self.__traced_errors.get(key)
            if not packets:
                self.__traced_errors[key] = copy.copy(value)
            else:
                packets["count"] += value["count"]

        for key, value in six.iteritems(stat.__traced_external_errors):
            packets = self.__traced_external_errors.get(key)
            if not packets:
                self.__traced_external_errors[key] = copy.copy(value)
            else:
                packets["count"] += value["count"]

    def merge_metric_packets(self, snapshot):
        """
        :param snapshot:
        :return:
        """
        for key, value in six.iteritems(snapshot.__time_packets):
            packets = self.__time_packets.get(key)
            if not packets:
                self.__time_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(snapshot.__apdex_packets):
            packets = self.__apdex_packets.get(key)
            if not packets:
                self.__apdex_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        for key, value in six.iteritems(snapshot.__action_packets):
            packets = self.__action_packets.get(key)
            if not packets:
                self.__action_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        # TODO: think more about the background task
        for key, value in six.iteritems(snapshot.__traced_errors):
            packets = self.__traced_errors.get(key)
            if not packets:
                self.__traced_errors[key] = copy.copy(value)
            else:
                packets["item"][-3] += value["count"]

        for key, value in six.iteritems(snapshot.__traced_external_errors):
            packets = self.__traced_external_errors.get(key)
            if not packets:
                self.__traced_external_errors[key] = copy.copy(value)
            else:
                packets["item"][-3] += value["count"]

        # generate general data
        for key, value in six.iteritems(snapshot.__general_packets):
            packets = self.__general_packets.get(key)
            if not packets:
                self.__general_packets[key] = copy.copy(value)
            else:
                packets.merge_packets(value)

        # for action trace
        top_n = self.__settings.action_tracer.top_n
        for key, value in six.iteritems(snapshot.__slow_action):

            if key not in self.__slow_action:
                self.__slow_action[key] = value
                break

            slow_actons = self.__slow_action.get(key)
            # although the target action trace value is `list`, but it only has 1 element in one metric.
            if len(slow_actons) > top_n:
                console.debug("The action trace is reach the top(%s), action(%s) is ignored.", top_n, key)
                break
            slow_actons.extend(value)

        # for slow sql
        max_sql = self.__settings.slow_sql_count
        for key, value in six.iteritems(snapshot.__slow_sql_packets):
            if len(self.__slow_sql_packets) > max_sql:
                console.debug("The slow sql trace count is reach the top.")
                continue

            slow_sql = self.__slow_sql_packets.get(key)
            if not slow_sql:
                self.__slow_sql_packets[key] = value
            else:
                slow_sql.merge_packets(value)

        # for quantile
        if 0 != len(self.__settings.quantile):
            action, duration = snapshot.__quantile_data
            if action not in self.__quantile:
                self.__quantile[action] = QuantileP2(self.__settings.quantile)

            self.__quantile[action].add(duration)

    def reset_metric_packets(self):
        """
        :return:
        """
        self.__time_packets = {}  # component
        self.__apdex_packets = {}
        self.__action_packets = {}
        self.__general_packets = {}

        self.__traced_errors = {}
        self.__traced_external_errors = {}
        self.__slow_action = {}
        self.__slow_sql_packets = {}

        self.__quantile = {}

    def packets_snapshot(self):
        """
        :return:
        """
        stat = copy.copy(self)

        self.__time_packets = {}
        self.__action_packets = {}
        self.__apdex_packets = {}
        self.__traced_errors = {}
        self.__traced_external_errors = {}
        self.__general_packets = {}
        self.__slow_action = {}
        self.__slow_sql_packets = {}
        self.__quantile = {}

        return stat

    # just for upload performance package
    # strip to 5 parts
    def component_metrics(self, metric_name_ids):
        """
        :return:
        """
        result = []

        for key, value in six.iteritems(self.__time_packets):
            extend_metrics = key[0].split("|")
            if len(extend_metrics) == 1:
                upload_key = {"name": key[0], "parent": key[1]}
                upload_key_str = '%s:%s' % (key[0], key[1])
                upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
                result.append([upload_key, value])
            elif len(extend_metrics) == 3:
                upload_key = {"name": extend_metrics[0], "parent": key[1], "calleeId": extend_metrics[1],
                              "calleeName": extend_metrics[2]}
                result.append([upload_key, value])

        self.__time_packets = {}
        return result

    def apdex_data(self, metric_name_ids):
        """
        :return:
        """
        result = []

        for key, value in six.iteritems(self.__apdex_packets):
            upload_key = {"name": key[0]}
            upload_key_str = '%s' % key[0]
            upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
            result.append([upload_key, value])

        # reset the data if returned for upload.
        self.__apdex_packets = {}
        return result

    def action_metrics(self, metric_name_ids):
        """
        :return:
        """
        result = []
        for key, value in six.iteritems(self.__action_packets):
            upload_key = {"name": key[0]}
            upload_key_str = '%s' % key[0]
            upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]

            if 0 != len(self.__settings.quantile):
                result.append([upload_key, value, self.__quantile[upload_key_str].markers])
            else:
                result.append([upload_key, value])

        # reset the data if returned for upload.
        self.__action_packets = {}
        return result

    def error_packets(self, metric_name_ids):
        """stat the error trace metric for performance
        :return:
        """
        external_error, web_action_error = 'External', 'WebAction'
        error_types = [external_error, web_action_error]
        error_count = {
            "Errors/Count/All": 0,
            "Errors/Count/AllWeb": 0,
            "Errors/Count/AllBackground": 0
        }

        def parse_error_trace(traced_data):
            for error_filter_key, error in six.iteritems(traced_data):
                error_count["Errors/Count/All"] += error["count"]

                if error["tracker_type"] in error_types:
                    error_count["Errors/Count/AllWeb"] += error["count"]

                    action_key = "Errors/Count/%s" % error_filter_key.split("_|")[0]
                    if action_key not in error_count:
                        error_count[action_key] = error["count"]
                    else:
                        error_count[action_key] += error["count"]

                    if error["tracker_type"] == external_error:
                        action_key = "Errors/Type:%s/%s" % (error["status_code"], error_filter_key.split("_|")[0])
                        if action_key not in error_count:
                            error_count[action_key] = error["count"]
                        else:
                            error_count[action_key] += error["count"]
                else:
                    error_count["Errors/Count/AllBackground"] += 1

        parse_error_trace(self.__traced_errors)
        parse_error_trace(self.__traced_external_errors)

        stat_value = []
        for key, value in six.iteritems(error_count):
            upload_key = {"name": key}
            upload_key_str = '%s' % key
            upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
            stat_value.append([upload_key, [value]])

        return stat_value

    # stat for error trace data
    # rely to the basic error trace data structure
    def error_trace_data(self):
        """
        :return:
        """
        rtv = [error["item"] for error in six.itervalues(self.__traced_errors)]
        self.__traced_errors = {}

        return rtv

    def external_error_data(self):
        rtv = [error["item"] for error in six.itervalues(self.__traced_external_errors)]
        self.__traced_external_errors = {}

        return rtv

    def general_trace_metric(self, metric_name_ids):
        """
        :return:
        """
        result = []
        for key, value in six.iteritems(self.__general_packets):
            extend_keys = key[0].split("|")
            if len(extend_keys) == 1:
                upload_key = {"name": key[0]}
                upload_key_str = '%s' % key[0]
                upload_key = upload_key if upload_key_str not in metric_name_ids else metric_name_ids[upload_key_str]
                result.append([upload_key, value])
            elif len(extend_keys) == 3:
                # do not replace the metric with id.
                upload_key = {"name": extend_keys[0], "calleeId": extend_keys[1], "calleeName": extend_keys[2]}
                result.append([upload_key, value])

        self.__general_packets = {}
        return result

    def action_trace_data(self):
        """
        :return:
        """
        if not self.__slow_action:
            return []

        trace_data = []

        for traces in six.itervalues(self.__slow_action):
            for trace in traces:
                trace_data.append(trace[0])

        return {"type": "actionTraceData", "actionTraces": trace_data}

    def slow_sql_data(self):
        """
        :return:
        """
        if not self.__slow_sql_packets:
            return []

        result = {"type": "sqlTraceData", "sqlTraces": []}
        maximum = self.__settings.slow_sql_count
        slow_sql_nodes = sorted(six.itervalues(self.__slow_sql_packets), key=lambda x: x.max_call_time)[-maximum:]

        for node in slow_sql_nodes:
            explain_plan = node.slow_sql_node.explain_plan
            params = {"explainPlan": explain_plan if explain_plan else {}, "stacktrace": []}

            if node.slow_sql_node.stack_trace:
                for line in node.slow_sql_node.stack_trace:
                    if len(line) >= 4 and 'tingyun' not in line[0]:
                        params['stacktrace'].append("%s(%s:%s)" % (line[2], line[0], line[1]))

            result['sqlTraces'].append([node.slow_sql_node.start_time, node.slow_sql_node.path,
                                        node.slow_sql_node.metric, node.slow_sql_node.request_uri,
                                        node.slow_sql_node.formatted, node.call_count,
                                        node.total_call_time, node.max_call_time,
                                        node.min_call_time, str(params)])

        self.__slow_sql_packets = {}
        return result



# -*- coding: utf-8 -*-

"""this module define the  mongodb trace data node

"""

from collections import namedtuple
from tingyun.logistics.attribution import TimeMetric, node_start_time, node_end_time

_MONGO_NODE = namedtuple("_MONGO_NODE", ['schema', 'method', 'children', 'start_time', 'end_time', 'duration',
                                         'exclusive', 'host', 'port'])


class MongoNode(_MONGO_NODE):
    """
    """
    def parse_db(self, schema):
        """从schema表中解析出数据库、collection、operation
        :param schema:
        :return: (db, collection, operation)
        """
        db, collection, operation = "Unknown", "Unknown", "Unknown"

        parts = str(schema).split(".", 1)
        if 2 != len(parts):
            return db, collection, operation
        else:
            db = parts[0]

        parts = str(parts[1]).split("/", 1)
        if len(parts) >= 2:
            collection = parts[0]
            operation = parts[1]
        else:
            collection = parts[0] if len(parts) >= 1 else "Unknown"

        return db, collection, operation

    def time_metrics(self, root, parent):
        """
        :param root:
        :param parent:
        :return:
        """
        method = str(self.method).upper()
        db, collection, _ = self.parse_db(self.schema)

        name = 'GENERAL/MongoDB/%s:%s%%2F%s/All' % (self.host, self.port, db)
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = 'GENERAL/MongoDB/NULL/All'
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        if root.type == 'WebAction':
            name = "GENERAL/MongoDB/NULL/AllWeb"
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)
        else:
            name = "GENERAL/MongoDB/NULL/AllBackgound"
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = "GENERAL/MongoDB/%s:%s%%2F%s%%2F%s/%s" % (self.host, self.port, db, collection, method)
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = "MongoDB/%s:%s%%2F%s%%2F%s/%s" % (self.host, self.port, db, collection, method)
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

    def trace_node(self, root):
        """
        :param root:
        :return:
        """
        method = str(self.method).upper()
        params = {}
        children = []
        call_count = 1
        class_name = ""
        method_name = root.name
        root.trace_node_count += 1
        start_time = node_start_time(root, self)
        end_time = node_end_time(root, self)
        db, collection, _ = self.parse_db(self.schema)
        metric_name = "MongoDB/%s:%s%%2F%s%%2F%s/%s" % (self.host, self.port, db, collection, method)
        call_url = metric_name

        return [start_time, end_time, metric_name, "", call_count, class_name, method, params, children]



# -*- coding: utf-8 -*-

import sys
import logging
from tingyun.armoury.ammunition.tornado.name_parser import object_name
from tingyun.armoury.ammunition.function_tracker import FunctionTracker
from tingyun.armoury.ammunition.tornado.wrappers import wrap_function_wrapper, function_wrapper
from tingyun.armoury.ammunition.tornado_4.utils import TrackerTransferContext
from tingyun.armoury.ammunition.tornado_4.utils import obtain_tracker_from_request, record_exception

console = logging.getLogger(__name__)


@function_wrapper
def trace_handler_execute(wrapped, instance, args, kwargs):
    """
    """
    handler = instance
    request = handler.request

    if not request:
        console.warning("No request instance got from handler. this maybe agent potential design issues. "
                        "if this continue, please report to us.")
        return wrapped(*args, **kwargs)

    tracker = obtain_tracker_from_request(request)
    if not tracker:
        return wrapped(*args, **kwargs)

    if request.method not in handler.SUPPORTED_METHODS:
        name = object_name(wrapped)
    else:
        name = object_name(getattr(handler, request.method.lower()))

    tracker.set_tracker_name(name, priority=3)
    with TrackerTransferContext(tracker):
        return wrapped(*args, **kwargs)


@function_wrapper
def _do_method_trace(wrapped, instance, args, kwargs):

    tracker = obtain_tracker_from_request(instance.request)
    if not tracker:
        return wrapped(*args, **kwargs)

    with FunctionTracker(tracker, name=object_name(wrapped)):
        return wrapped(*args, **kwargs)


@function_wrapper
def _do_http_method_trace(wrapped, instance, args, kwargs):
    """所有从用户代码开始的性能采集要从内存中获取tracker对象，而不是request对象中获取，
    后续所有的性能捕获，都需要上下文管理器修饰
    """
    tracker = obtain_tracker_from_request(instance.request)
    if not tracker:
        return wrapped(*args, **kwargs)

    tracker.set_tracker_name(object_name(wrapped), priority=3)
    with FunctionTracker(tracker, name=object_name(wrapped)):
        return wrapped(*args, **kwargs)


def trace_handler_init(wrapped, instance, args, kwargs):
    """
    """
    def parse_class_name(method):
        """parse the class name of the method belong to
        """
        for cls in method.__self__.__class__.__mro__:
            if method.__name__ in cls.__dict__:
                return cls.__name__

        return "Unknown"

    def parse_request_method(application, request, *args, **kwargs):
        """接口参考RequestHandler.__init__(self, application, request, **kwargs)
        :return:
        """
        return str(request.method).lower()

    result = wrapped(*args, **kwargs)

    # 正常情况下tracker应该附加在instance.request上，若没有初始化成功，直接放弃打包
    tracker = obtain_tracker_from_request(instance.request)
    if not tracker:
        console.debug("tracker is not initialized before give up trace the function %s", object_name(wrapped))
        return wrapped(*args, **kwargs)

    # 由于是动态`临时`打包， 只捕获请求方法的性能即可
    method = parse_request_method(*args, **kwargs)
    func = getattr(instance, method, None)
    if func and not hasattr(func, "_previous_object"):
        setattr(instance, method, _do_http_method_trace(func))

    # prepare & on_finish & _handle_request_exception 多半是在用户代码实现，
    # 所以需要在实际初始化的时候进行处理. 如果这些函数使用了coroutine，则他们可能已经被打包过了，
    # 所以需要检查下，方式多次打包
    if parse_class_name(instance.prepare) != 'RequestHandler' and not hasattr(instance.prepare, "_previous_object"):
        instance.prepare = _do_method_trace(instance.prepare)

    if parse_class_name(instance.on_finish) != 'RequestHandler' and not hasattr(instance.on_finish, "_previous_object"):
        instance.on_finish = _do_method_trace(instance.on_finish)

    if _do_method_trace(instance._handle_request_exception) != 'RequestHandler' and \
            not hasattr(instance._handle_request_exception, "_previous_object"):
        instance._handle_request_exception = _do_method_trace(instance._handle_request_exception)

    #  防止用户覆盖此方法导致数据抓取异常，tracker无法进入线程队列
    instance._execute = trace_handler_execute(instance._execute)

    return result


def trace_handler_request_exception(wrapped, instance, args, kwargs):
    """
    """
    tracker = obtain_tracker_from_request(instance.request)

    with TrackerTransferContext(tracker):
        record_exception(sys.exc_info())
        return wrapped(*args, **kwargs)


def trace_handler_flush(wrapped, instance, args, kwargs):
    """数据被写入socket的时候获取响应状态码
    :return:
    """
    request = instance.request
    if not request:
        console.warning("No request instance got from handler, this maybe agent potential design issues.")
        return wrapped(*args, **kwargs)

    tracker = obtain_tracker_from_request(request)
    if not tracker:
        return wrapped(*args, **kwargs)

    tracker.deal_response('%s ok' % getattr(instance, '_status_code') or 500, {})
    return wrapped(*args, **kwargs)


def detect_web(module):
    """
    :param module:
    :return:
    """
    wrap_function_wrapper(module, "RequestHandler.__init__", trace_handler_init)
    wrap_function_wrapper(module, "RequestHandler._handle_request_exception", trace_handler_request_exception)
    wrap_function_wrapper(module, "RequestHandler.flush", trace_handler_flush)



"""define some wrapper for memcache client `pymemcache`
"""

import logging
from tingyun.armoury.ammunition.memcache_tracker import wrap_memcache_trace


console = logging.getLogger(__name__)
methods = ['add', 'append', 'cas', 'decr', 'delete', 'delete_many', 'delete_multi', 'flush_all', 'get',
           'get_many', 'get_multi', 'gets', 'gets_many', 'incr', 'prepend', 'replace', 'set', 'set_many', 'set_multi',
           'stats']


def detect_base_client(module):
    """
    :param module:
    :return:
    """
    def parse_connect_params(instance, *args, **kwargs):
        """
        :param instance:
        :param args:
        :param kwargs:
        :return:
        """
        host, port = "Unknown", "Unknown"

        if hasattr(instance, 'server') and 2 == len(instance.server):
            host, port = instance.server[0], instance.server[1]

        return host or 'Unknown', port or 'Unknown'

    for m in methods:
        if hasattr(module.Client, m):
            wrap_memcache_trace(module, 'Client.%s' % m, m, server=parse_connect_params)
            console.debug("Wrap the base client method %s", m)


def detect_pooled_client(module):
    """for Pooled client
    :param module:
    :return:
    """
    if not hasattr(module, "PooledClient"):
        console.info("Pooled Client not in this version.")
        return

    def parse_connect_params(instance, *args, **kwargs):
        """
        :param instance:
        :param args:
        :param kwargs:
        :return:
        """
        host, port = "Unknown", "Unknown"

        if hasattr(instance, 'server') and 2 == len(instance.server):
            host, port = instance.server[0], instance.server[1]

        return host, port

    for m in methods:
        if hasattr(module.PooledClient, m):
            wrap_memcache_trace(module, 'PooledClient.%s' % m, m, server=parse_connect_params)
            console.debug("Wrap the base pooled client method %s", m)


def detect_has_client(module):
    """this is a client for communicating with a cluster of memcached servers
    :param module:
    :return:
    """
    def parse_connect_params(instance, *args, **kwargs):
        """
        :param instance:
        :param args:
        :param kwargs:
        :return:
        """
        host, port = "Unknown", "Unknown"

        try:
            if hasattr(instance, 'clients') and 1 == len(instance.clients):
                key = instance.clients.keys()[0]
                host_port = str(key).split(":")
                host, port = host_port[0], host_port[1]
        except (AttributeError, IndexError) as err:
            console.warning("Parse host and port for pymongo hash client error %s", err)

        return host, port

    for m in methods:
        if hasattr(module.HashClient, m):
            wrap_memcache_trace(module, 'HashClient.%s' % m, m, server=parse_connect_params)
            console.debug("Wrap the hash client method %s", m)



from collections import namedtuple

from tingyun.logistics.attribution import TimeMetric, node_start_time, node_end_time


_FunctionNode = namedtuple('_FunctionNode', ['group', 'name', 'children', 'start_time', 'end_time', 'duration',
                                             'exclusive', 'params', 'stack_trace'])


class FunctionNode(_FunctionNode):
    """
    """
    def time_metrics(self, root, parent):
        """
        :param root:
        :param parent:
        :return:
        """
        name = 'Python/%s/%s' % (self.group, self.name)

        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        for child in self.children:
            for metric in child.time_metrics(root, self):
                yield metric

    def trace_node(self, root):
        """
        :param root: the root node of the trakcer
        :return: traced node
        """
        start_time = node_start_time(root, self)
        end_time = node_end_time(root, self)
        metric_name = 'Python/%s/%s' % (self.group, self.name)
        call_url = root.request_uri
        call_count = 1
        class_name = ""
        method_name = self.name
        params = {"sql": "", "explainPlan": {}, "stacktrace": []}
        children = []

        root.trace_node_count += 1
        for child in self.children:
            if root.trace_node_count > root.trace_node_limit:
                break

            children.append(child.trace_node(root))

        if self.stack_trace:
            for line in self.stack_trace:
                if len(line) >= 4 and 'tingyun' not in line[0]:
                    params['stacktrace'].append("%s(%s:%s)" % (line[2], line[0], line[1]))

        return [start_time, end_time, metric_name, call_url, call_count, class_name, method_name, params, children]



"""define some wrapper for memcache client `python-binary-memcache`
"""

import logging
from tingyun.armoury.ammunition.memcache_tracker import wrap_memcache_trace

console = logging.getLogger(__name__)
methods = ['add', 'cas', 'decr', 'delete', 'delete_multi', 'delete_multi', 'get',
           'get_multi', 'gets', 'incr', 'replace', 'set', 'set_multi',
           'stats']


def detect_client(module):
    """
    :param module:
    :return:
    """
    def parse_connect_params(instance, *args, **kwargs):
        """
        :param instance:
        :param args:
        :param kwargs:
        :return:
        """
        host, port = "Unknown", "Unknown"
        if hasattr(instance, '_servers') and len(instance._servers) > 0:
            try:
                host, port = instance._servers[0].host, instance._servers[0].port
            except AttributeError:
                pass

        return host, port

    for m in methods:
        if hasattr(module.Client, m):
            wrap_memcache_trace(module, 'Client.%s' % m, m, server=parse_connect_params)
            console.debug("Wrap the base client method %s", m)



# -*- coding: utf-8 -*-

"""define some detector for tornado4 web module
"""

import logging
import sys

from tingyun.armoury.ammunition.tornado_4.name_parser import object_name
from tingyun.armoury.ammunition.tornado_4.wrappers import wrap_function_wrapper, function_wrapper
from tingyun.armoury.ammunition.function_tracker import FunctionTracker
from tingyun.armoury.ammunition.tornado_4.utils import obtain_tracker_from_request
from tingyun.armoury.ammunition.tornado_4.utils import record_exception, TrackerTransferContext

console = logging.getLogger(__name__)


@function_wrapper
def _do_method_trace(wrapped, instance, args, kwargs):
    tracker = obtain_tracker_from_request(instance.request)
    if not tracker:
        return wrapped(*args, **kwargs)

    # 用户可能会重写RequestHandler中的_excute方法， 导致无法正确抓取metricName
    name = object_name(wrapped)
    http_methods = name.split(".")
    if len(http_methods) >= 2 and http_methods[-1] == instance.request.method.lower():
        tracker.set_tracker_name(name, priority=3)

    with FunctionTracker(tracker, name=name):
        return wrapped(*args, **kwargs)


def trace_handler_init(wrapped, instance, args, kwargs):
    """
    """
    def parse_class_name(method):
        """parse the class name of the method belong to
        """
        for cls in method.__self__.__class__.__mro__:
            if method.__name__ in cls.__dict__:
                return cls.__name__
        return None

    result = wrapped(*args, **kwargs)

    for method in instance.SUPPORTED_METHODS:
        method = str(method).lower()
        func = getattr(instance, method, None)

        # 如果使用coroutine装饰过，会被附加上_previous_object属性,那么则跳过
        if func and not hasattr(func, "_previous_object"):
            setattr(instance, method, _do_method_trace(func))

    if parse_class_name(instance.prepare) != 'RequestHandler':
        instance.prepare = _do_method_trace(instance.prepare)

    if parse_class_name(instance.on_finish) != 'RequestHandler':
        instance.on_finish = _do_method_trace(instance.on_finish)

    return result


def trace_handler_request_exception(wrapped, instance, args, kwargs):
    """
    """
    tracker = obtain_tracker_from_request(instance.request)

    with TrackerTransferContext(tracker):
        record_exception(sys.exc_info())
        return wrapped(*args, **kwargs)


def trace_handler_execute(wrapped, instance, args, kwargs):
    """
    """
    handler = instance
    request = handler.request

    if not request:
        console.warning("No request instance got from handler. this maybe agent potential design issues. "
                        "if this continue, please report to us.")
        return wrapped(*args, **kwargs)

    tracker = obtain_tracker_from_request(request)
    if not tracker:
        return wrapped(*args, **kwargs)

    if request.method not in handler.SUPPORTED_METHODS:
        name = object_name(wrapped)
    else:
        name = object_name(getattr(handler, request.method.lower()))

    tracker.set_tracker_name(name, priority=3)
    with TrackerTransferContext(tracker):
        return wrapped(*args, **kwargs)


def trace_handler_flush(wrapped, instance, args, kwargs):
    """
    :return:
    """
    request = instance.request
    if not request:
        console.warning("No request instance got from handler, this maybe agent potential design issues.")
        return wrapped(*args, **kwargs)

    tracker = obtain_tracker_from_request(request)
    if not tracker:
        return wrapped(*args, **kwargs)

    tracker.deal_response('%s ok' % getattr(instance, '_status_code') or 500, {})
    return wrapped(*args, **kwargs)


def detect_web(module):
    """
    :param module:
    :return:
    """
    wrap_function_wrapper(module, "RequestHandler._execute", trace_handler_execute)
    wrap_function_wrapper(module, "RequestHandler._handle_request_exception", trace_handler_request_exception)
    wrap_function_wrapper(module, "RequestHandler.__init__", trace_handler_init)
    wrap_function_wrapper(module, "RequestHandler.flush", trace_handler_flush)





"""Define this module for basic armory for detect mako

"""

from tingyun.armoury.ammunition.function_tracker import wrap_function_trace


def detect_template(module):
    """
    :param module:
    :return:
    """
    def template_name(template, text, filename, *args):
        """
        :param template:
        :param text:
        :param filename:
        :param args: compatible with _compile_text and _compile_module_file
        :return:
        """
        return filename

    def template_name_in_render(instance, *args, **kwargs):
        return getattr(instance, 'filename', "template")

    wrap_function_trace(module, '_compile_text', name=template_name, group='Template.compile')
    wrap_function_trace(module, '_compile_module_file', name=template_name, group='Template.compile')
    wrap_function_trace(module, 'Template.render', name=template_name_in_render, group='Template.render')



"""Define this module for basic armory for gevent

"""

from tingyun.armoury.trigger.wsgi_entrance import wsgi_app_wrapper_entrance
from tingyun.logistics.basic_wrapper import trace_in_function


def trace_wsgi_server(*args, **kwargs):
    def instance_parameters(instance, listener, application, *args, **kwargs):
        return instance, listener, application, args, kwargs

    self, listener, application, _args, _kwargs = instance_parameters(*args, **kwargs)

    application = wsgi_app_wrapper_entrance(application)

    _args = (self, listener, application) + _args

    return _args, _kwargs


def detect_pywsgi(module):
    """
    """
    trace_in_function(module, 'WSGIServer.__init__', trace_wsgi_server)


# this module will be deprecated after gevent v1.0.3
def detect_wsgi(module):
    """
    """
    trace_in_function(module, 'WSGIServer.__init__', trace_wsgi_server)




"""define module for tornado wsgi entrance. but not support tornado core functions

"""

import logging
from tingyun.armoury.trigger.wsgi_entrance import wsgi_application_wrapper

console = logging.getLogger(__name__)


def detect_wsgi_entrance(module):
    """
    :param module:
    :return:
    """
    try:
        import tornado
        version = tornado.version
    except Exception:
        version = "xx"

    if hasattr(module, "WSGIAdapter"):
        wsgi_application_wrapper(module.WSGIAdapter, '__call__', ('Tornado', version))
    else:
        console.warning("WSGIAdapter not exist in tornado 3.x, and we are not support 3.x version now.")




# Note: alpha/beta version should not be delivered to customer
#       only ('alpha', 'beta', 'rc', 'final') used
#       {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
VERSION = (1, 3, 0, 'final', 0)


def get_version():
    """
    :return:
    """
    from tingyun.version import get_version
    return get_version(VERSION)



# -*- coding: utf-8 -*-

import json
import logging

from collections import namedtuple
from tingyun.logistics.attribution import TimeMetric, ApdexMetric, TracedError, node_start_time, node_end_time
from tingyun.logistics.attribution import TracedExternalError


console = logging.getLogger(__name__)
_TrackerNode = namedtuple('_TrackerNode', ['type', 'group', 'name', 'start_time', 'end_time', 'request_uri',
                                           'duration', 'http_status', 'exclusive', 'children', 'path', "errors",
                                           "apdex_t", "request_params", "custom_params", "thread_name", "trace_data",
                                           "referer", "slow_sql", "queque_time", "trace_guid", 'trace_id',
                                           "external_error", "trace_interval", 'call_external_id'])


class TrackerNode(_TrackerNode):
    """hold the tracker trace data for collect

    """
    def time_metrics(self):
        """
        :return:
        """
        yield TimeMetric(name=self.path, scope=self.path, duration=self.duration, exclusive=self.exclusive)

        queque_metric = 'GENERAL/WebFrontend/NULL/QueueTime'
        yield TimeMetric(name=queque_metric, scope=queque_metric, duration=self.queque_time, exclusive=self.queque_time)

        for child in self.children:
            for metric in child.time_metrics(self, self):
                yield metric

    def quantile(self):
        """
        :return:
        """
        return self.path, self.duration

    def action_metrics(self):
        """
        :return: the full tracker metric of the top
        """
        if self.type != "WebAction":
            return

        if self.http_status > 400:
            console.debug("Abnormal status code %s with uri %s", self.http_status, self.request_uri)
            return

        yield TimeMetric(name=self.path, scope="", duration=self.duration, exclusive=self.exclusive)

    def apdex_metrics(self):
        """
        :return:
        """
        if self.type != "WebAction":
            console.debug("get apdex with none webaction %s", self.type)
            return

        satisfying = 0
        tolerating = 0
        frustrating = 0

        # status code large than 400 but not 401, we think it frustrating
        if (self.http_status >= 400 and self.http_status != 401) or self.errors:
            frustrating = 1
        else:
            if self.duration <= self.apdex_t:
                satisfying = 1
            elif self.duration <= 4 * self.apdex_t:
                tolerating = 1
            else:
                frustrating = 1

        name = self.path.replace("WebAction", "Apdex")
        yield ApdexMetric(name=name, satisfying=satisfying, tolerating=tolerating, frustrating=frustrating,
                          apdex_t=self.apdex_t)

    def traced_error(self):
        """yield the traced errors according to protocol
        :return:
        """
        error_type_cache = []

        for error in self.errors:
            error_filter_key = "%s_|%s_|%s_|%s" % (self.path, error.http_status, error.error_class_name, error.message)
            if error_filter_key in error_type_cache:
                console.debug("Duplicate error `%s` for uri %s", error.error_class_name, self.request_uri)
                continue

            error_item = [error.error_time, self.path, self.http_status, error.error_class_name, error.message,
                          1, self.request_uri]
            stack_detail = []
            for line in error.stack_trace:
                if len(line) >= 4 and 'tingyun' not in line[0]:
                    stack_detail.append("%s(%s:%s)" % (line[2], line[0], line[1]))

            error_params = {"params": {"threadName": error.thread_name, "httpStatus": self.http_status,
                                       "referer": error.referer},
                            "requestParams": error.request_params,
                            "stacktrace": stack_detail
                            }

            error_item.append(json.dumps(error_params))
            error_type_cache.append(error_filter_key)

            yield TracedError(error_filter_key=error_filter_key, tracker_type=error.tracker_type, trace_data=error_item)

    def trace_node(self, root):
        """
        :param root: the root node of the tracker
        :return: traced node
        """
        start_time = node_start_time(root, self)
        end_time = node_end_time(root, self)
        metric_name = self.path
        call_url = self.request_uri
        call_count = 1
        class_name = ""
        method_name = self.name
        params = {}
        children = []

        root.trace_node_count += 1
        for child in self.children:
            if root.trace_node_count > root.trace_node_limit:
                break

            children.append(child.trace_node(root))

        return [start_time, end_time, metric_name, call_url, call_count, class_name, method_name, params, children]

    def slow_action_trace(self, limit, threshold):
        """ the main interface to data engine
        :param limit: the maximum limitation of the nodes
        :param threshold: this value is dynamic from server. then pass from packager
        :return:
        """
        self.trace_node_limit = limit
        self.trace_node_count = 0
        start_time = int(self.start_time)
        duration = self.duration
        trace_id = self.trace_id
        trace_guid = self.trace_guid

        # 被调用方如果产生慢过程，调用方，需要强制触发该慢过程
        trace_data = self.trace_data if self.trace_data else {}
        if not trace_data.get("tr", 0) and duration < threshold:
            return None

        trace_node = self.trace_node(self)
        custom_params = {"httpStatus": self.http_status, "threadName": self.thread_name, "referer": self.referer,
                         }
        custom_params.update(self.custom_params)
        slow_trace = [start_time, self.request_params, custom_params, trace_node]

        return [start_time, duration, self.path, self.request_uri, json.dumps(slow_trace),
                trace_id, trace_guid]

    def slow_sql_nodes(self):
        """
        :return:
        """
        for item in self.slow_sql:
            yield item.slow_sql_node(self)

    def traced_external_error(self):
        """
        :return:
        """
        for error in self.external_error:
            metric_name = 'External/%s/%s' % (error.url.replace("/", "%2F"), error.module_name)
            error_item = [error.error_time, metric_name, error.status_code,
                          error.error_class_name, 1, self.path]

            stack_detail = []
            for line in error.stack_trace:
                if len(line) >= 4 and 'tingyun' not in line[0]:
                    stack_detail.append("%s(%s:%s)" % (line[2], line[0], line[1]))

            error_params = {"params": {"threadName": error.thread_name, "httpStatus": self.http_status},
                            "requestParams": error.request_params, "stacktrace": stack_detail}
            error_item.append(json.dumps(error_params))

            error_filter_key = "%s_|%s_|%s" % (metric_name, self.http_status, error.error_class_name)
            yield TracedExternalError(error_filter_key=error_filter_key, trace_data=error_item,
                                      tracker_type=error.tracker_type, status_code=self.http_status)




"""this module is implement the function detector for web2py

"""

import sys
from tingyun.armoury.ammunition.function_tracker import FunctionTracker
from tingyun.logistics.basic_wrapper import FunctionWrapper, wrap_object, import_module
from tingyun.logistics.object_name import callable_name
from tingyun.armoury.ammunition.tracker import current_tracker


def uncaught_exception_wrapper(wrapped):
    """
    :param wrapped:
    :return:
    """
    def wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()

        def _wrapper(request, response, session):
            """
            """
            name = '%s.%s.%s.%s' % (request.application, request.controller, request.function, request.extension)
            tracker.set_tracker_name(name=name, group='web2py')
            wrapped(request, response, session)

        if tracker is None:
            return wrapped(*args, **kwargs)

        HTTP = import_module('gluon.http').HTTP
        try:
            return _wrapper(*args, **kwargs)
        except HTTP:
            raise
        except:
            tracker.record_exception(*sys.exc_info())
            raise

    return FunctionWrapper(wrapped, wrapper)


def trace_serve_controller(module, object_path):
    """all application errors in web2py will raised in this function.
    :param module:
    :param object_path:
    :return:
    """
    wrap_object(module, object_path, uncaught_exception_wrapper)




# -*- coding: utf-8 -*-

import logging
import threading
import weakref

from tingyun.battlefield.tracer import Tracer
from tingyun.battlefield.proxy import proxy_instance
from tingyun.armoury.ammunition.tracker import current_tracker

console = logging.getLogger(__name__)
ty_local = threading.local()


def is_websocket(request):
    """Now web ignore websocket feature. it's not in our current conversation.
    :param request: HTTP request header
    :return:
    """
    if request.headers.get('Upgrade', '').lower() == 'websocket':
        console.debug("current request is web socket, agent will ignore this tracker.")
        return True

    return False


def request_environ(environ, request):
    """the environ in the default environ may not include our need environment variable when apply to wsgi application
    :param environ: returned environment from the container request.
    :param request:
    :return:
    """
    # we need this the build the request information. but not included in the environment.
    ret = dict(environ)
    ret["REQUEST_URI"] = request.uri
    ret["HTTP_REFERER"] = request.headers.get('Referer', '')
    ret['HTTP_X_QUEUE_START'] = request.headers.get("X-Queue-Start")

    if not environ:
        ret['PATH_INFO'] = request.path
        ret['SCRIPT_NAME'] = ""
        ret['QUERY_STRING'] = request.query

    return ret


def obtain_current_tracker():
    """ Get the tracker from tracer cache.
    :return:
    """
    return current_tracker()


def drop_current_tracer():
    """ separate the tracker from agent thread
    :return:
    """
    old_tracker = obtain_current_tracker()
    if old_tracker is not None:
        old_tracker.drop_tracker()
    return old_tracker


def replace_current_tracer(new_tracker):
    """ Set the current tracer into agent thread and return the old one.
    :param new_tracker:
    :return:
    """
    old_tracker = drop_current_tracer()
    if new_tracker:
        new_tracker.save_tracker()

    return old_tracker


def obtain_request(tracker):
    """obtain the request object from tracker. which is attached at began
    :param tracker:
    :return:
    """
    # its' weakref
    request = getattr(tracker, "_nb_request", None)
    return request() if request else None


class TrackerTransferContext(object):
    def __init__(self, tracer):
        self.tracer = tracer

    def __enter__(self):
        self.old_tracer = replace_current_tracer(self.tracer)

    def __exit__(self, exc_type, exc_value, traceback):
        replace_current_tracer(self.old_tracer)


def record_exception(exc_info, tracker=None):
    """
    :return:
    """
    import tornado.web
    exc = exc_info[0]

    # Not an error so we just return.
    if exc is tornado.web.Finish:
        return

    tracker = tracker or obtain_current_tracker()
    if tracker:
        tracker.record_exception(*exc_info)


def obtain_tracker_from_request(request):
    """
    :param request: httpserver request
    :return:
    """
    tracker = getattr(request, '_nb_tracker', None)
    if not tracker:
        console.debug("No tracker found in request[%s], if this continues, please report to us.", request)

    return tracker


def generate_tracer(request, framework='Tornado'):
    """
    :param request: the http request for client
    :param framework: framework name
    :return:
    """
    old_tracker = drop_current_tracer()
    tracker = Tracer(proxy_instance(), request_environ({}, request), framework)
    if not tracker.enabled:
        return None

    tracker.start_work()
    drop_current_tracer()  # drop it from thread cache
    tracker._nb_request = weakref.ref(request)
    request._nb_tracker = tracker

    replace_current_tracer(old_tracker)

    return tracker


def finish_tracker(tracker, exc_type=None, exc_val=None, exc_tb=None):
    if not tracker:
        console.error("Tracker is None, shit maybe some trace strategy issue. if this continues. please report to us.")
        return

    if getattr(tracker, "_is_finalized", None):
        console.error("This errors maybe caused by some agent logic trace error. if this continues, please report to"
                      " us for further investigation.")
        return

    old_tracker = replace_current_tracer(tracker)
    try:
        tracker.finish_work(exc_type, exc_val, exc_tb, async=True)
    finally:
        setattr(tracker, "_is_finalized", True)

        if getattr(tracker, "_nb_request", None):
            setattr(tracker, "_nb_request", None)

        request = getattr(tracker, "_nb_request", None)
        if hasattr(request, "_nb_tracker"):
            setattr(request, "_nb_tracker", None)

        # put back the previous tracker to tracker cache.
        if tracker != old_tracker:
            replace_current_tracer(old_tracker)




""" Prepare all need strategic materials to take advantage for the war
"""
from tingyun.embattle.inspection import take_control


def initialize(config_file=None):
    """ Can be triggered by application frame.
    :param config_file: config file for corps
    :return:
    """
    status = take_control(config_file).execute()
    return status




"""define this module for detect django-piston package
"""

import logging
from tingyun.logistics.object_name import callable_name
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.logistics.basic_wrapper import wrap_function_wrapper
from tingyun.armoury.ammunition.function_tracker import FunctionTracker


console = logging.getLogger(__name__)


def trace_resource_call(wrapped, instance, args, kwargs):
    """
    :param wrapped:
    :param instance:
    :param args:
    :param kwargs:
    :return:
    """
    method_mapping = {'GET': 'read', 'POST': 'create', 'PUT': 'update', 'DELETE': 'delete'}

    def parse_request(request, *args, **kwargs):
        return request

    tracker = current_tracker()
    if not tracker:
        return wrapped(*args, **kwargs)

    request = parse_request(*args, **kwargs)
    method = request.method.upper()

    # if invalid request method, use the origin request method
    # And using high priority to set the metric name.
    name = "%s.%s" % (callable_name(instance.handler), method_mapping.get(method, method))
    tracker.set_tracker_name(name=name, priority=5)
    with FunctionTracker(tracker, name):
        return wrapped(*args, **kwargs)


def detect_piston_resource(module):
    """
    :param module:
    :return:
    """
    wrap_function_wrapper(module, "Resource.__call__", trace_resource_call)



# -*- coding: utf-8 -*-

"""define tracker for tracer use.

"""

import logging
import time
import random
import sys
import traceback

from tingyun.armoury.ammunition.timer import Timer
from tingyun.battlefield.knapsack import knapsack
from tingyun.logistics.warehouse.database_node import DatabaseNode
from tingyun.logistics.warehouse.tracker_node import TrackerNode
from tingyun.logistics.warehouse.error_node import ErrorNode, ExternalErrorNode
from tingyun.packages import six

console = logging.getLogger(__name__)


class Tracker(object):
    """Tracker, the battle entrance.
    """
    def __init__(self, proxy, enabled=None, framework="Python"):
        """
        :param proxy: communication proxy instance to controller.
        :return:
        """
        self.framework = framework
        self.proxy = proxy
        self.enabled = False
        self.rum_enabled = True  # current transaction switch to control auto insert browser agent.
        self._settings = None

        self.thread_id = knapsack().current_thread_id()

        self.background_task = False
        self.start_time = 0
        self.end_time = 0
        self.trace_interval = 0  # 距离上一次
        self.queque_start = 0
        self._queque_time = 0
        self.trace_node = []
        self.async_func_trace_time = 0  # millisecond

        self._errors = []
        self.external_error = []
        self._custom_params = {}
        self._slow_sql_nodes = []

        self.http_method = "GET"
        self.request_uri = None
        self.http_status = 500
        self.request_params = {}
        self.header_params = {}
        self.cookie_params = {}
        self._priority = 0
        self._group = None
        self._name = None
        self.apdex = 0
        self._frozen_path = None

        self.stack_trace_count = 0
        self.explain_plan_count = 0

        self.thread_name = "Unknown"
        self.referer = ""

        # 作为调用者存储的自己的或者被调者返回的信息
        self._trace_guid = ""  # 被调用时，若产生慢过程生成该id用于标识此次慢过程
        self._tingyun_id = ""  # 该实例自身的id
        self._called_traced_data = ""
        self.called_external_id = ''

        # 做为被调用者，存储的调用者信息
        self.call_trace_id = ""  # 调用方产生的trace id
        self.call_external_id = ""  # 调用方产生的外部调用id
        self.call_tingyun_id = ""  # 调用方发出的听云id，包含账号id
        self.call_req_id = ""  # 调用方请求标识

        # 0 indicate it's not called the service.
        self.db_time = 0
        self.external_time = 0
        self.redis_time = 0
        self.memcache_time = 0
        self.mongo_time = 0

        # set the global switch of agent.  the return application settings is download from server witch is merged
        # with the local settings on server
        # agent always be enabled if enabled in local settings. just not upload data when disabled by server.
        global_settings = proxy.global_settings
        if global_settings.enabled:
            if enabled or (enabled is None and proxy.enabled):
                self._settings = proxy.settings
                if not self._settings:
                    self.proxy.activate()
                    self._settings = proxy.settings

                if self._settings:
                    self.enabled = True

        if self.enabled:
            self._tingyun_id = self._settings.tingyunIdSecret

    def __enter__(self):
        """
        :return:
        """
        if not self.enabled:
            return self

        try:
            self.save_tracker()
        except Exception as _:
            console.fatal("Fatal error when save tracker. if this continues. please contact us for further \
                           investigation")
            raise

        self.start_time = time.time()
        if self.proxy.previous_time:
            trace_interval = int(self.start_time * 1000) - self.proxy.previous_time
            self.trace_interval = trace_interval
        else:
            self.trace_interval = 0

        self.trace_node.append(Timer(None))

    def __exit__(self, exc_type, exc_val, exc_tb, async=False):
        """
        :param exc_type:
        :param exc_val:
        :param exc_tb:
        :param async: in tornado, the tracker is not saved in thread cache
        :return:
        """
        if not self.enabled:
            return

        if not async:
            try:
                self.drop_tracker()
            except Exception as err:
                console.exception("error detail %s", err)
                raise

        if not self.is_uri_captured(self.request_uri):
            console.debug("ignore the uri %s", self.request_uri)
            return

        if exc_type is not None and exc_val is not None and exc_tb is not None:
            self.record_exception(exc_type, exc_val, exc_tb)

        # 结束时间以应用最后一个字节写入时间为准，在框架结束的地方可能已经处理好了
        self.end_time = time.time()
        self.proxy.previous_time = int(self.end_time * 1000)
        duration = self.duration

        root_node = self.trace_node.pop()
        children = root_node.children
        exclusive = duration + root_node.exclusive

        tracker_type = "WebAction" if not self.background_task else "BackgroundAction"
        group = self._group or ("Python" if self.background_task else "Uri")
        request_params = self.filter_params(self.request_params)
        uri = self.request_uri

        # replace the specified metric apdex_t
        apdex_t = self._settings.apdex_t
        path = self.path
        if path in self._settings.action_apdex:
            apdex_t = self._settings.action_apdex.get(path)

        self.finalize_tracker_data()

        node = TrackerNode(type=tracker_type, group=group, name=self._name, start_time=self.start_time,
                           end_time=self.end_time, request_uri=uri, duration=duration, thread_name=self.thread_name,
                           http_status=self.http_status, exclusive=exclusive, children=tuple(children),
                           path=self.path, errors=self._errors, apdex_t=apdex_t, queque_time=self.queque_time,
                           custom_params=self._custom_params, request_params=request_params,
                           trace_id=self.call_trace_id, call_external_id=self.call_external_id,
                           referer=self.referer, slow_sql=self._slow_sql_nodes, trace_guid=self.generate_trace_guid(),
                           trace_data=self._called_traced_data, external_error=self.external_error,
                           trace_interval=self.trace_interval)

        self.proxy.record_tracker(node)

    def record_exception(self, exc=None, value=None, tb=None, params=None, tracker_type="WebAction",
                         ignore_errors=[]):
        """ record the exception for trace
        :return:
        """
        if not self._settings or not self._settings.error_collector.enabled:
            return

        if exc is None and value is None and tb is None:
            exc, value, tb = sys.exc_info()

        if exc is None or value is None or tb is None:
            console.warning("None exception is got. skip it now. %s, %s, %s", exc, value, tb)
            return

        if self.http_status in self._settings.error_collector.ignored_status_codes:
            console.debug("ignore the status code %s", self.http_status)
            return

        # 'True' - ignore the error.
        # 'False'- record the error.
        # ignore status code and maximum error number filter will be done in data engine because of voiding repeat count
        # method ignore_errors() is used to detect the the status code which is can not captured
        if callable(ignore_errors):
            should_ignore = ignore_errors(exc, value, tb, self._settings.error_collector.ignored_status_codes)
            if should_ignore:
                return

        # think more about error occurred before deal the status code.
        if self.http_status in self._settings.error_collector.ignored_status_codes:
            console.debug("record_exception: ignore  error collector status code")
            return

        module = value.__class__.__module__
        name = value.__class__.__name__
        fullname = '%s:%s' % (module, name) if module else name

        request_params = self.filter_params(self.request_params)
        if params:
            custom_params = dict(request_params)
            custom_params.update(params)
        else:
            custom_params = dict(request_params)

        try:
            message = str(value)
        except Exception as _:
            try:
                # Assume JSON encoding can handle unicode.
                message = six.text_type(value)
            except Exception as _:
                message = '<unprintable %s object>' % type(value).__name__

        stack_trace = traceback.extract_tb(tb)
        node = ErrorNode(error_time=int(time.time()), http_status=self.http_status, error_class_name=fullname,
                         uri=self.request_uri, thread_name=self.thread_name, message=message,
                         stack_trace=stack_trace, request_params=custom_params, tracker_type=tracker_type,
                         referer=self.referer)

        self._errors.append(node)

    def record_external_error(self, url, error_code, http_status=0, _exception=None, request_params=None,
                              tracker_type="External", module_name="Unknown"):
        """
        :return:
        """
        ignored_status = [401, ]
        if http_status in ignored_status or int(http_status) < 400:
            console.info("Agent caught http status code %s, ignore it now.", http_status)
            return

        status_code = error_code or http_status
        error_class_name = ""
        stack_trace = ""

        if _exception:
            exc, value, tb = _exception
            module = value.__class__.__module__
            name = value.__class__.__name__
            error_class_name = '%s:%s' % (module or "Unknown", name or "Unknown")
            stack_trace = traceback.extract_tb(tb)

        node = ExternalErrorNode(error_time=int(time.time()), status_code=status_code, thread_name=self.thread_name,
                                 url=url, error_class_name=error_class_name, http_status=http_status,
                                 stack_trace=stack_trace, request_params=request_params, tracker_type=tracker_type,
                                 module_name=module_name)

        self.external_error.append(node)

    def finalize_tracker_data(self):
        """处理追踪器中需要最好采集的数据，如不能回传的跨应用数据
        :return:
        """
        if not self.call_tingyun_id or not self.settings:
            return None

        if self.duration < self.settings.action_tracer.action_threshold:
            return None

        app_id = self.call_tingyun_id[self.call_tingyun_id.find('|') + 1: ]
        code_time = self.duration - self.queque_time - self.db_time - self.external_time - self.redis_time - \
                    self.memcache_time - self.mongo_time
        data = {
            "applicationId": app_id,
            "transactionId": self.call_trace_id,
            "externalId": self.call_external_id,
            "time": {
                "duration": self.duration,
                "qu": self.queque_time,
                "db": self.db_time,
                "ex": self.external_time,
                "rds": self.redis_time,
                "mc": self.memcache_time,
                "mon": self.mongo_time,
                "code": 0 if code_time < 0 else code_time
            }
        }

        self._custom_params['message.wait'] = self.trace_interval
        self._custom_params['entryTrace'] = data

    @property
    def path(self):
        """the call tree path
        """
        if self._frozen_path:
            return self._frozen_path

        tracker_type = "WebAction" if not self.background_task else "BackgroundAction"
        name = self._name or "Undefined"

        try:
            named_metric = self.settings.naming.naming_web_action(self.http_method, self.request_uri,
                                                                  self.request_params,
                                                                  self.header_params, self.cookie_params)
        except Exception as err:
            console.error("Error occurred when parsing naming rules. %s", err)
            named_metric = None

        if named_metric:
            path = '%s/%s' % (tracker_type, self.url_encode(named_metric))
            return path

        uri_params_captured = self._settings.web_action_uri_params_captured.get(self.request_uri, '')
        if self.request_uri in self._settings.web_action_uri_params_captured:
            match_param = ''
            for actual_param in self.request_params:
                if actual_param in uri_params_captured:
                    match_param += "&%s=%s" % (actual_param, self.request_params[actual_param])
            match_param = match_param.replace("&", "?", 1)
            self.request_uri = "%s%s" % (self.request_uri, match_param)

        if not self._settings.auto_action_naming:
            path = "%s/%s/%s" % (tracker_type, self.framework, self.url_encode(self.request_uri))
        else:
            path = '%s/%s/%s' % (tracker_type, self.framework, name)

        return path

    def url_encode(self, uri):
        """
        :param uri:
        :return:
        """
        encoded_url = "index"
        if not uri or uri == "/":
            return encoded_url

        # drop the uri first /
        merged_uri = self.merge_urls(uri.replace('/', '', 1))
        encoded_url = merged_uri.replace("/", "%2F")

        return encoded_url

    def merge_urls(self, url):
        """
        :param url:
        :return:
        """
        if not self.settings.urls_merge or self.settings.auto_action_naming:
            return url

        def is_digit(world):
            if str(world).isdigit():
                return '*'

            return world

        url_re = '/'.join([is_digit(s) for s in url.split('/')])
        url_segments = []
        pre_is_digit = False
        is_param = False

        for s in str(url_re):
            # skip the query parameters
            if s == "?":
                is_param = True

            if is_param:
                url_segments.append(s)
                continue

            # current and the last is digit. replace it.
            if str(s).isdigit() and pre_is_digit:
                url_segments[-1] = '*'

            elif str(s).isdigit() and not pre_is_digit:
                pre_is_digit = True
                url_segments.append(s)

            else:
                url_segments.append(s)
                pre_is_digit = False

        return ''.join(url_segments)

    def set_tracker_name(self, name, group="Function", priority=None):
        """used to consist metric identification
        """
        if priority is None:
            return

        if priority is not None and priority < self._priority:
            return

        if isinstance(name, bytes):
            name = name.decode('Latin-1')

        self._priority = priority
        self._group = group
        self._name = name

    def save_tracker(self):
        knapsack().save_tracker(self)

    def drop_tracker(self):
        if not self.enabled:
            return self

        knapsack().drop_tracker(self)

    def process_database_node(self, node):
        """record the database node
        :param node: database node
        :return:
        """
        if type(node) is not DatabaseNode:
            return

        if not self._settings.action_tracer.enabled:
            return

        if not self._settings.action_tracer.slow_sql:
            return

        if self._settings.action_tracer.record_sql == 'off':
            return

        if node.duration < self._settings.action_tracer.slow_sql_threshold:
            return

        self._slow_sql_nodes.append(node)

    def push_node(self, node):
        self.trace_node.append(node)

    def pop_node(self, node):
        """return parent trace node
        """
        last = self.trace_node.pop()
        assert last == node
        parent = self.trace_node[-1]

        return parent

    def current_node(self):
        """
        """
        if self.trace_node:
            return self.trace_node[-1]

    def is_uri_captured(self, uri):
        """check the uri is captured or not
        :param uri: uri of the request
        :return: True, if it's need to captured. else return False
        """
        # capture all url
        if not self._settings.urls_captured:
            return True

        for p in self._settings.urls_captured:
            if p and p.match(uri):
                return True

        return False

    def filter_params(self, params):
        """filter the parameters
        :param params: the dict parameters
        :return: filtered parameters
        """
        result = {}

        if not params:
            return result

        if not self._settings.action_tracer.enabled or not self._settings.capture_params:
            return result

        for key in params:
            if key not in self._settings.ignored_params:
                result[key] = params[key]

        return result

    @property
    def name(self):
        return self._name

    @property
    def group(self):
        return self._group

    @property
    def settings(self):
        return self._settings

    @property
    def duration(self):
        """
        :return:
        """
        end_time = self.end_time
        if not self.end_time:
            end_time = time.time()

        return int((end_time - self.start_time) * 1000)

    def generate_trace_guid(self):
        """
        :return:
        """
        duration = self.duration
        if duration < self.settings.action_tracer.action_threshold:
            return ""

        if self._trace_guid:
            return self._trace_guid

        guid = '%s%s' % (time.time(), random.random())
        guid = guid.replace('.', '')
        self._trace_guid = guid[:30]
        return self._trace_guid

    def generate_trace_id(self):
        """ call_trace_id 由调用方生成，若被调方又作为调用方，则复用调用方的call_trace_id
        :return:
        """

        if self.call_trace_id:
            return self.call_trace_id

        trace_id = '%s%s' % (time.time(), random.random())
        trace_id = trace_id.replace('.', '')
        self.call_trace_id = trace_id[:16]
        return self.call_trace_id

    @property
    def queque_time(self):
        """
        :return:
        """
        if not self._queque_time:
            _queque_time = int(self.start_time * 1000.0 - self.queque_start)
            self._queque_time = _queque_time if _queque_time > 0 and self.queque_start != 0 else 0

        return self._queque_time


def current_tracker():
    """return current tracker in the thread
    :return:
    """
    return knapsack().current_tracker()




"""transportation base tool for upload

"""

import time
import logging
import sys
import json
import zlib

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from tingyun.logistics.exceptions import DiscardDataForRequest, RetryDataForRequest, InvalidLicenseException
from tingyun.logistics.exceptions import InvalidDataTokenException, OutOfDateConfigException, ServerIsUnavailable
from tingyun.config.settings import global_settings
from tingyun.packages import six, requests

console = logging.getLogger(__name__)

# 10kbi
COMPRESS_MINIMUM_SIZE = 10 * 1024
USER_AGENT = ' NBS Newlens Agent/%s (python %s; %s)' % (global_settings().agent_version, sys.version.split()[0],
                                                        sys.platform)


def parse_proxy():
    """
    :return: {'http': 'xxx', 'https': 'xxx'}
    """
    settings = global_settings()

    # Proxy schema set in configure file,then use it
    # Else use `ssl` configure to decide 'http' or 'https'
    host = settings.proxy_host or ''
    port = settings.proxy_port or ''
    user = settings.proxy_user or ''
    scheme = settings.proxy_scheme or 'http'
    password = settings.proxy_pwd or ''

    # we support `http://user:password@host` format, so just checkout host
    if not host:
        return

    # urlparse will return <scheme>://<netloc>/<path>
    # (scheme, netloc, path, params, query, fragment)
    sections = urlparse.urlparse(host)
    if (not sections.scheme or sections.scheme not in ('http', 'https')) and not port:
        return

    path = ''
    netloc = host
    if sections.scheme:  # config the  host with <scheme>://<netloc>/<path> in settings
        scheme = sections.scheme
        netloc = sections.netloc
        path = sections.path
    elif sections.path:  # urlparse parse the netloc to path
        netloc = sections.path

    if port:
        netloc = "%s:%s" % (netloc, port)

    if user:
        if password:
            netloc = '%s:%s@%s' % (user, password, netloc)
        else:
            netloc = '%s@%s' % (user, netloc)

    full_url = '%s://%s%s' % (scheme, netloc, path)
    return {'http': full_url, 'https': full_url}


def transmitter(session, url, action, payload={}, param={}, audit_mode=False):
    """
    :param session: the request session to server
    :param url: the address witch data send to
    :param action: the send actually intention
    :param payload: request data
    :return: None
    """
    console.debug("Send request with url %s, action %s param %s", url, action, param)
    settings = global_settings()
    start_time = time.time()
    headers = {
        "User-Agent": USER_AGENT,
        "Content-Type": "application/octet-stream",
        "connection": "close",
        "Accept-Encoding": "deflate",
    }

    try:
        data = json.dumps(payload)
    except Exception as err:
        console.error("Encoding json for payload failed, url %s action %s param %s payload %s err %s",
                      url, action, param, payload, err)
        raise DiscardDataForRequest(str(sys.exc_info()[1]))

    if len(data) > COMPRESS_MINIMUM_SIZE:
        headers['Content-Encoding'] = 'deflate'
        level = (len(data) < 2000000) and 1 or 9
        data = zlib.compress(six.b(data), level)

    auto_close_session = False
    if not session:
        session = requests.session()
        auto_close_session = True

    content = ""
    try:
        # because of the root license use the sha1 encryption, the version of the certifi new than 2015.4.28
        # which used sha256 encryption, will cause the ssl verify failed, so no mater what kind of agent work
        # mode(private deploy or public use), we do not verify the ssl license now.

        verify_ssl = False if settings.port else settings.verify_certification
        timeout = settings.data_report_timeout
        proxy = parse_proxy()
        ret = session.post(url, data=data, params=param, headers=headers, timeout=timeout, verify=verify_ssl,
                           proxies=proxy)
        content = ret.content
    except requests.RequestException:
        console.error('Agent server is not attachable. if the error continues, please report to networkbench.'
                      'The error raised was %s. response content is %s.', sys.exc_info()[1], content)
        raise RetryDataForRequest(str(sys.exc_info()[1]))
    finally:
        duration = time.time() - start_time
        if auto_close_session:
            session.close()

    if audit_mode:
        console.info("Use %ss to upload data return value %r", duration, content)

    if ret.status_code == 400:
        console.error("Bad request has been submitted for url %s, please report this to us, thank u.", url)
        raise DiscardDataForRequest()
    elif ret.status_code == 503:
        console.error("Agent server is unavailable. This can be a transient issue because of the server or our core"
                      " application being restarted. if this error continues, please report to us. thank u.")
        raise DiscardDataForRequest()
    elif ret.status_code == 502:
        console.error("Agent server error, our engineer has caution this error, thanks for your support")
        raise ServerIsUnavailable("service unavailable, get status code 502")
    elif ret.status_code != 200:
        console.warning("We got none 200 status code %s, this maybe some network/server error, if this error continues,"
                        "please report to us . thanks for your support. return content %s", ret.status_code, ret)
        raise DiscardDataForRequest()

    try:
        if six.PY3:
            content = content.decode('UTF-8')

        result = json.loads(content)
    except Exception as err:
        console.warning("Decoding data for Json error. please contact us for further investigation. %s", err)
        raise DiscardDataForRequest(str(sys.exc_info()[1]))
    else:
        # successful exchange with server
        if result["status"] == "success":
            return result["result"] if "result" in result else []

    console.info("get unexpected return,  there maybe some issue. %s", result)
    server_status = int(result["result"]["errorCode"])

    if server_status == 460:
        console.warning("Invalid license key, Please contact to networkbench for more help.")
        raise InvalidLicenseException("Invalid license key")
    elif server_status == 462:
        console.warning("Invalid data format, maybe something get wrong when json encoding.")
        raise DiscardDataForRequest(content)
    elif server_status == 461:
        console.warning("Invalid data token, if this error continues, please report to networkbench support for further"
                        " investigation")
        raise InvalidDataTokenException()
    elif server_status == -1:
        console.warning("Agent server error, our engineer has caution this error, thanks for your support.")
        raise ServerIsUnavailable()
    elif server_status == 470:
        console.info("Configuration is out of date, server configuration will be obtain again")
        raise OutOfDateConfigException()

    return []





def get_version(version):
    """Returns a PEP 386-compliant version number from VERSION.

    """
    version_mapping = {'alpha': 'a', 'beta': 'b', 'rc': 'c'}
    assert len(version) == 5
    assert version[3] in ('alpha', 'beta', 'rc', 'final')

    parts = 3
    main = '.'.join(str(x) for x in version[:parts])
    sub = ''
    if version[3] != 'final':
        sub = version_mapping[version[3]] + str(version[4])

    return str(main + sub)


# -*- coding: utf-8 -*-

import logging

import time
from tingyun.logistics.basic_wrapper import wrap_function_wrapper
from tingyun.armoury.ammunition.tornado.util import finish_tracker, obtain_tracker_from_request

console = logging.getLogger(__name__)


def trace_tracker_export(wrapped, instance, args, kwargs):
    """
    """
    # request将以两种方式存在,一种是常规的tornado.Application  另外一种是WSGIContainer模式下,此时为非代理模式访问
    request = instance.request or instance.delegate.request
    if not request:
        console.warning("Errors, when finish the tracer when finish the tracer, this maybe some issue of design,"
                        "if this continue, pleae report to us, thank u.")
        return wrapped(*args, **kwargs)

    tracker = obtain_tracker_from_request(request)
    if not tracker:
        console.debug("Can not get trace from request, tracer maybe finished before.")
        return wrapped(*args, **kwargs)

    try:
        return wrapped(*args, **kwargs)
    finally:
        if getattr(request, "_finish_time", 0):
            tracker.end_time = getattr(request, "_finish_time", int(time.time()))

        finish_tracker(tracker)


def detect_http_server(module):
    """请求出口,  由于采用计数方式计算整个事务的生命周期,该出口调用时, 整个请求并未结束
    :param module:
    :return:
    """
    wrap_function_wrapper(module, "_ServerRequestAdapter.finish", trace_tracker_export)
    wrap_function_wrapper(module, "_ServerRequestAdapter.on_connection_close", trace_tracker_export)



# -*- coding: utf-8 -*-

"""提供上层接口计算分位数
"""

import logging
import copy

console = logging.getLogger(__name__)


class QuantileP2(object):
    """
    """
    def __init__(self, quartile_list):
        """分位数列表， 最多4个
        :param quartile_list: list
        """
        assert isinstance(quartile_list, list)
        assert len(quartile_list) > 0

        self.quartile_list = quartile_list
        self.quartile_list.sort()

        self.markers_y = [0.0 for _ in range(len(self.quartile_list) * 2 + 3)]
        self.markers_x = None
        self.count = 0
        self.p2_n = []

        self.init_marks()

    @property
    def markers(self):
        """
        :return:
        """
        if self.count < len(self.markers_y):
            result = [0 for _ in range(self.count)]
            markers = [0 for _ in range(len(self.markers_y))]
            pw_q_copy = copy.deepcopy(self.markers_y)
            pw_q_copy.sort()

            i = len(pw_q_copy) - self.count
            j = 0
            while i < len(pw_q_copy):
                result[j] = pw_q_copy[i]
                i += 1
                j += 1

            for i in range(len(pw_q_copy)):
                markers[i] = result[int(round((self.count - 1) * i * 1.0 / (len(pw_q_copy) - 1)))]

            return markers

        return self.markers_y

    def init_marks(self):
        """
        :return:
        """
        quartile_count = len(self.quartile_list)
        marker_count = quartile_count * 2 + 3

        self.markers_x = [0.0 for _ in range(marker_count)]
        self.p2_n = [0 for _ in range(len(self.markers_y))]

        for i in range(quartile_count):
            marker = self.quartile_list[i]
            self.markers_x[i * 2 + 1] = (marker + self.markers_x[i * 2]) / 2.0
            self.markers_x[i * 2 + 2] = marker

        self.markers_x[marker_count - 2] = (1 + self.quartile_list[quartile_count - 1]) / 2.0
        self.markers_x[marker_count - 1] = 1.0

        for i in range(marker_count):
            self.p2_n[i] = i

    def add(self, v):
        """
        :param v:
        :return:
        """
        if v is None:
            return

        obs_idx = self.count
        self.count += 1

        if obs_idx < len(self.markers_y):
            self.markers_y[obs_idx] = v

            if obs_idx == len(self.markers_y) - 1:
                self.markers_y.sort()
        else:
            k = self.binary_search(self.markers_y, v)
            if k < 0:
                k = -(k + 1)

            if 0 == k:
                self.markers_y[0] = v
                k = 1
            elif k == len(self.markers_y):
                k = len(self.markers_y) - 1
                self.markers_y[k] = v

            for i in range(k, len(self.p2_n)):
                self.p2_n[i] += 1

            for i in range(1, len(self.markers_y) - 1):
                n_ = self.markers_x[i] * obs_idx
                di = n_ - self.p2_n[i]

                if (di-1.0 >= 0.000001 and self.p2_n[i + 1] - self.p2_n[i] > 1) or \
                        (di+1.0 <= 0.000001 and self.p2_n[i - 1] - self.p2_n[i] < -1):
                    d = -1 if di < 0 else 1
                    qi_ = self.quad_pred(d, i)

                    if qi_ < self.markers_y[i - 1] or qi_ > self.markers_y[i + 1]:
                        qi_ = self.line_pred(d, i)

                    self.markers_y[i] = qi_
                    self.p2_n[i] += d

    def binary_search(self, data, v):
        """
        :param data:
        :param v:
        :return:
        """
        low, high = 0, len(data) - 1

        while low <= high:
            mid = (low + high) >> 1

            if data[mid] < v:
                low = mid + 1
            elif data[mid] > v:
                high = mid - 1
            else:
                if data[mid] == v:
                    return mid
                elif data[mid] < v:
                    low = mid + 1
                else:
                    high = mid - 1

        return -(low + 1)

    def quad_pred(self, d, i):
        """
        :param d:
        :param i:
        :return:
        """
        qi = self.markers_y[i]
        qip1 = self.markers_y[i + 1]
        qim1 = self.markers_y[i - 1]
        ni = self.p2_n[i]
        nip1 = self.p2_n[i + 1]
        nim1 = self.p2_n[i - 1]

        a = (ni - nim1 + d) * (qip1 - qi) / (nip1 - ni)
        b = (nip1 - ni - d) * (qi - qim1) / (ni - nim1)
        return qi + (d * (a + b)) / (nip1 - nim1)

    def line_pred(self, d, i):
        """
        :param d:
        :param i:
        :return:
        """
        qi = self.markers_y[i]
        qipd = self.markers_y[i + d]
        ni = self.p2_n[i]
        nipd = self.p2_n[i + d]

        return qi + d * (qipd - qi) / (nipd - ni)



# -*- coding: utf-8 -*-

import inspect

from tingyun.packages.wrapt.wrappers import wrap_object
from tingyun.packages.wrapt.wrappers import (FunctionWrapper as _FunctionWrapper, ObjectProxy as _ObjectProxy,
                                             BoundFunctionWrapper as _BoundFunctionWrapper)


class _ObjectWrapperBase(object):

    def __setattr__(self, name, value):
        if name.startswith('_nb_'):
            name = name.replace('_nb_', '_self_', 1)
            setattr(self, name, value)
        else:
            _ObjectProxy.__setattr__(self, name, value)

    def __getattr__(self, name):
        if name.startswith('_nb_'):
            name = name.replace('_nb_', '_self_', 1)
            return getattr(self, name)
        else:
            return _ObjectProxy.__getattr__(self, name)

    def __delattr__(self, name):
        if name.startswith('_nb_'):
            name = name.replace('_nb_', '_self_', 1)
            delattr(self, name)
        else:
            _ObjectProxy.__delattr__(self, name)

    @property
    def _next_object(self):
        return self.__wrapped__

    @property
    def _previous_object(self):
        try:
            return self._self_previous_object
        except AttributeError:
            self._self_previous_object = getattr(self.__wrapped__, '_previous_object', self.__wrapped__)
            return self._self_previous_object


class _NbBoundFunctionWrapper(_ObjectWrapperBase, _BoundFunctionWrapper):
    pass


class FunctionWrapper(_ObjectWrapperBase, _FunctionWrapper):
    __bound_function_wrapper__ = _NbBoundFunctionWrapper


def function_wrapper(wrapper):
    def _wrapper(wrapped, instance, args, kwargs):
        target_wrapped = args[0]
        if instance is None:
            target_wrapper = wrapper
        elif inspect.isclass(instance):
            target_wrapper = wrapper.__get__(None, instance)
        else:
            target_wrapper = wrapper.__get__(instance, type(instance))

        return FunctionWrapper(target_wrapped, target_wrapper)

    return FunctionWrapper(wrapper, _wrapper)


def wrap_function_wrapper(module, name, wrapper):
    return wrap_object(module, name, FunctionWrapper, (wrapper,))





"""Define this module for basic armory for django

"""
from tingyun.armoury.trigger.wsgi_entrance import wsgi_application_wrapper
from tingyun.armoury.ammunition.django_tracker import trace_middleware_wrapper, trace_view_dispatch
from tingyun.armoury.ammunition.django_tracker import trace_template_block_render, trace_django_template
from tingyun.armoury.ammunition.django_tracker import should_ignore, trace_django_urlresolvers
from tingyun.armoury.ammunition.django_tracker import trace_urlresolvers_resolve_xxx, trace_uncaught_exception
from tingyun.armoury.ammunition.function_tracker import wrap_function_trace
from tingyun.armoury.ammunition.error_tracker import wrap_error_trace


# Wrap the WSGI application entry point for django.
def detect_wsgi_entrance(module):
    import django

    framework = 'Django'
    version = django.get_version()

    wsgi_application_wrapper(module.WSGIHandler, "__call__", (framework, version))
    trace_uncaught_exception(module.WSGIHandler, "handle_uncaught_exception")


# Post import hooks for modules.
def detect_middleware(module):
    """process the http/response middleware
    :param module: the module need to detect
    :return:
    """
    trace_middleware_wrapper(module, 'BaseHandler.load_middleware')


def detect_views_dispatch(module):
    """process the views info
    :param module: the views module
    :return:
    """
    trace_view_dispatch(module, "View.dispatch")


def detect_urlresolvers(module):
    """process the url
    :param module:
    :return:
    """
    # for grab the resolver map error. detect the function name mapping to views, this is base call for resolver
    wrap_error_trace(module, 'get_callable', ignore_errors=should_ignore)

    # wrap the regex url mapping to views. get_callable is used inline
    module.RegexURLResolver.resolve = trace_django_urlresolvers(module.RegexURLResolver.resolve)

    if hasattr(module.RegexURLResolver, 'resolve403'):
        module.RegexURLResolver.resolve403 = trace_urlresolvers_resolve_xxx(module.RegexURLResolver.resolve403, priority=3)

    if hasattr(module.RegexURLResolver, 'resolve404'):
        module.RegexURLResolver.resolve404 = trace_urlresolvers_resolve_xxx(module.RegexURLResolver.resolve404, priority=3)

    if hasattr(module.RegexURLResolver, 'resolve500'):
        module.RegexURLResolver.resolve500 = trace_urlresolvers_resolve_xxx(module.RegexURLResolver.resolve500, priority=1)


def detect_template_block_render(module):
    """
    :param module:
    :return:
    """
    trace_template_block_render(module, "BlockNode.render")


def detect_django_template(module):
    """ detect the template render
    :param module:
    :return:
    """
    if hasattr(module.Template, '_render'):
        trace_django_template(module, 'Template._render')

    if hasattr(module.Template, 'render'):
        trace_django_template(module, 'Template.render')


def detect_http_multipartparser(module):
    """ detect for file upload
    :param module:
    :return:
    """
    wrap_function_trace(module, 'MultiPartParser.parse')


def detect_core_mail(module):
    """
    :param module:
    :return:
    """
    wrap_function_trace(module, 'mail_admins')
    wrap_function_trace(module, 'mail_managers')
    wrap_function_trace(module, 'send_mail')


def detect_core_mail_message(module):
    """
    :param module:
    :return:
    """
    wrap_function_trace(module, 'EmailMessage.send')




"""

"""

from tingyun.armoury.ammunition.memcache_tracker import wrap_memcache_trace

# option: {object_path: $object_path, command: $command}
memcache_attr = {
    "add": {"path": "Client.add", "command": "add"},
    "append": {"path": "Client.append", "command": "append"},
    "cas": {"path": "Client.cas", "command": "cas"},
    "decr": {"path": "Client.decr", "command": "decr"},
    "delete": {"path": "Client.delete", "command": "delete"},
    "delete_multi": {"path": "Client.delete_multi", "command": "delete"},
    "get": {"path": "Client.get", "command": "get"},
    "gets": {"path": "Client.gets", "command": "get"},
    "get_multi": {"path": "Client.get_multi", "command": "get"},
    "incr": {"path": "Client.incr", "command": "incr"},
    "prepend": {"path": "Client.prepend", "command": "prepend"},
    "replace": {"path": "Client.replace", "command": "replace"},
    "set": {"path": "Client.set", "command": "set"},
    "set_multi": {"path": "Client.set_multi", "command": "set"},
}


def detect(module):
    def parse_connect_param(instance, *args, **kwargs):
        """
        :return:
        """
        host, port = "Unknown", "Unknown"
        if hasattr(instance, 'servers') and len(instance.servers) > 0:
            try:
                host, port = instance.servers[0].ip, instance.servers[0].port
            except AttributeError:
                pass

        return host, port

    for attr in memcache_attr:
        if hasattr(module.Client, attr):
            wrap_memcache_trace(module, memcache_attr[attr]['path'], attr, server=parse_connect_param)




"""this module define the sampler controller for sampler the data.
"""
import logging

console = logging.getLogger(__name__)


class SamplerController(object):
    """
    """
    def __init__(self, sampler, *args):
        """
        :param sampler:
        :param args:
        :return:
        """
        self.name = args[0]
        self.sampler = sampler
        self.args = args
        self.instance = None

    def start(self):
        """
        :return:
        """
        if self.instance is None:
            self.instance = self.sampler(self.args)

        if hasattr(self.instance, 'start'):
            self.instance.start()
            return

    def stop(self):
        if hasattr(self.instance, 'stop'):
            self.instance.stop()
        else:
            self.instance = None

    def metrics(self, *args):
        if self.instance is None:
            return []

        return self.instance(*args)



# -*- coding: utf-8 -*-

"""
"""

import logging
from collections import namedtuple
from tingyun.logistics.warehouse.dbapi_tools import sql_parser
from tingyun.logistics.attribution import node_start_time, node_end_time, TimeMetric
from tingyun.config.settings import global_settings

_SlowSqlNode = namedtuple('_SlowSqlNode', ['duration', 'path', 'request_uri', 'sql', 'sql_format', 'metric', 'dbapi',
                                           'stack_trace', 'connect_params', 'cursor_params', 'execute_params',
                                           "start_time"])
_DatabaseNode = namedtuple('_DatabaseNode', ['dbapi', 'sql', 'children', 'start_time', 'end_time', 'duration',
                                             'exclusive', 'stack_trace', 'sql_format', 'connect_params',
                                             'cursor_params', 'execute_params', "dbtype", 'host', 'port', 'db_name'])
console = logging.getLogger(__name__)


class SlowSqlNode(_SlowSqlNode):
    """
    """

    def __new__(cls, *args, **kwargs):
        node = _SlowSqlNode.__new__(cls, *args, **kwargs)
        node.parser = sql_parser(node.sql, node.dbapi)
        return node

    @property
    def operation(self):
        return self.parser.operation

    @property
    def formatted(self):
        return self.parser.formatted(self.sql_format)

    @property
    def identifier(self):
        return self.parser.identifier

    @property
    def explain_plan(self):
        return self.parser.explain_plan(self.connect_params, self.cursor_params, self.execute_params)


class DatabaseNode(_DatabaseNode):
    def __new__(cls, *args, **kwargs):
        node = _DatabaseNode.__new__(cls, *args, **kwargs)
        node.parser = sql_parser(node.sql, node.dbapi)

        return node

    @property
    def operation(self):
        return self.parser.operation

    @property
    def table(self):
        return self.parser.table

    @property
    def formatted(self):
        return self.parser.formatted(self.sql_format)

    @property
    def explain_plan(self):
        return self.parser.explain_plan(self.connect_params, self.cursor_params, self.execute_params)

    def time_metrics(self, root, parent):
        """GENERAL 用于统计GENERAL中的一些组件的性能数据(通用汇总数据)
            非GENERAL将组成component数据
        :param root:
        :param parent:
        :return:
        """
        dbtype = r' %s' % self.dbtype
        host, port, db_name = self.host, self.port, self.db_name

        name = "GENERAL/Database%s/%s:%s%%2F%s/All" % (dbtype, host, port, db_name)
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = 'GENERAL/Database%s/NULL/All' % dbtype
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        if root.type == 'WebAction':
            name = "GENERAL/Database%s/NULL/AllWeb" % dbtype
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)
        else:
            name = "GENERAL/Database%s/NULL/AllBackgound" % dbtype
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        operation = str(self.operation).upper() or "CALL"

        if self.table:
            name = "GENERAL/Database%s/%s:%s%%2F%s%%2F%s/%s" % (dbtype, host, port, db_name, self.table, operation)
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

            name = "Database%s/%s:%s%%2F%s%%2F%s/%s" % (dbtype, host, port, db_name, self.table, operation)
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)
        else:
            name = "GENERAL/Database%s/%s:%s%%2F%s%%2FUnknown/%s" % (dbtype, host, port, db_name, operation)
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

            name = "Database%s/%s:%s%%2F%s%%2FUnknown/%s" % (dbtype, host, port, db_name, operation)
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = "GENERAL/Database%s/NULL/%s" % (dbtype, operation or "CALL")
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

    def trace_node(self, root):
        """
        :param root: the root node of the tracker
        :return:
        """
        dbtype = r" %s" % self.dbtype
        params = {"sql": "", "explainPlan": {}, "stacktrace": []}
        children = []
        call_count = 1
        class_name = ""
        root.trace_node_count += 1
        start_time = node_start_time(root, self)
        end_time = node_end_time(root, self)
        operation = str(self.operation).upper() or 'CALL'
        method_name = '%s.%s' % (self.dbtype, operation)
        metric_name = "Database%s/%s:%s%%2F%s%%2FUnknown/%s" % (dbtype, self.host, self.port, self.db_name, operation)
        call_url = ""
        host, port, db_name = self.host, self.port, self.db_name

        if self.table:
            metric_name = "Database %s/%s:%s%%2F%s%%2F%s/%s" % (dbtype, host, port, db_name, self.table, operation)
        else:
            console.debug("Can not get table for operate `%s` to `%s`", operation, dbtype)

        if self.formatted:
            # Note, use local setting only.
            _settings = global_settings()
            params['sql'] = self.formatted

            if _settings.action_tracer.log_sql:
                console.info("Log sql is opened. sql upload is disabled, sql sentence is %s", self.formatted)
                params['sql'] = ""

            if self.explain_plan:
                params['explainPlan'] = self.explain_plan

            if self.stack_trace:
                for line in self.stack_trace:
                    if len(line) >= 4 and 'tingyun' not in line[0]:
                        params['stacktrace'].append("%s(%s:%s)" % (line[2], line[0], line[1]))

        return [start_time, end_time, metric_name, call_url, call_count, class_name, method_name, params, children]

    def slow_sql_node(self, root):
        """
        :return:
        """
        dbtype = r" %s" % self.dbtype
        request_uri = root.request_uri.replace("%2F", "/")
        operation = str(self.operation).upper() or 'CALL'
        host, port, db_name = self.host, self.port, self.db_name
        metric_name = "Database%s/%s:%s%%2F%s%%2FUnknown/%s" % (dbtype, host, port, db_name, operation)

        if self.table:
            metric_name = "Database %s/%s:%s%%2F%s%%2F%s/%s" % (dbtype, host, port, db_name, self.table, operation)
        else:
            console.debug("Can not get table for operate `%s` to `%s`", operation, dbtype)

        return SlowSqlNode(duration=self.duration, path=root.path, request_uri=request_uri, metric=metric_name,
                           start_time=int(self.start_time), sql=self.sql, sql_format=self.sql_format, dbapi=self.dbapi,
                           stack_trace=self.stack_trace, connect_params=self.connect_params,
                           cursor_params=self.cursor_params, execute_params=self.execute_params)




"""Define this module for openstack nova model

"""

from tingyun.armoury.trigger.wsgi_entrance import wsgi_application_wrapper, TARGET_NOVA_APP


def resource_wsgi_entrance(module):
    """
    :param module:
    :return:
    """
    version = 'xx'
    try:
        import nova
        version = getattr(nova, "__version__", 'xx')
    except Exception as _:
        pass

    wsgi_application_wrapper(module.Resource, '__call__', ('nova', version, TARGET_NOVA_APP))



import logging

"""used to mapping the settings
"""


ENV_CONFIG_FILE = "TING_YUN_CONFIG_FILE"
CONSTANCE_OUT_DATE_CONFIG = -815
CONSTANCE_INVALID_DATA_TOKEN = CONSTANCE_OUT_DATE_CONFIG
CONSTANCE_DISCARD_DATA = -816
CONSTANCE_SERVER_UNAVAILABLE = -817
CONSTANCE_RETRY_DATA = -818
CONSTANCE_HARVEST_ERROR = -819
CONSTANCE_LICENSE_INVALID = -820
CONSTANCE_SESSION_NOT_ACTIVED = -830
CONSTANCE_INVALID_LICENSE_KEY = -831

MAX_RETRY_UPLOAD_FAILED_TIMES = 4  # four minutes

_LOG_LEVEL = {
    'CRITICAL': logging.CRITICAL,
    'ERROR': logging.ERROR,
    'WARNING': logging.WARNING,
    'INFO': logging.INFO,
    'DEBUG': logging.DEBUG,
    'NOTSET': logging.NOTSET,

    'FATAL': logging.FATAL,
    'WARN': logging.WARN
}

BOOL_MAPPING = {
    "ON": True,
    "OFF": False,
    "TRUE": True,
    "FALSE": False
}


def map_log_level(s):
    level = None
    try:
        return _LOG_LEVEL["%s" % str(s).upper()]
    except Exception as _:
        pass

    return level


# multiple name is not support
def map_app_name(s):
    return s or None


def map_include_environ(s):
    return str(s).split()


def map_key_words(v):
    """
    :param v:
    :return:
    """
    if isinstance(v, bool):
        return v

    v = str(v).upper()
    if v in BOOL_MAPPING:
        return BOOL_MAPPING[v]
    else:
        # invalid value, so handled in settings
        return None


CONFIG_ITEM = [
    {"section": "tingyun", "key": "license_key", "mapper": None},
    {"section": "tingyun", "key": "enabled", "mapper": map_key_words},
    {"section": "tingyun", "key": "app_name", "mapper": map_app_name},
    {"section": "tingyun", "key": "audit_mode", "mapper": map_key_words},
    {"section": "tingyun", "key": "auto_action_naming", "mapper": map_key_words},
    {"section": "tingyun", "key": "ssl", "mapper": map_key_words},
    {"section": "tingyun", "key": "action_tracer.log_sql", "mapper": map_key_words},
    {"section": "tingyun", "key": "log_file", "mapper": None},
    {"section": "tingyun", "key": "log_level", "mapper": map_log_level},
]



# -*- coding: utf-8 -*-

import logging
from collections import namedtuple
from tingyun.logistics.attribution import TimeMetric, node_start_time, node_end_time


console = logging.getLogger(__name__)
_MQNode = namedtuple('_MQNode', ['vendor', 'name_type', 'host', 'port', 'byte', 'name', 'headers', 'children', 'role',
                                 'wrapped_info', 'start_time', 'end_time', 'duration', 'exclusive', 'external_id'])


class MQNode(_MQNode):
    """
    """
    def time_metrics(self, root, parent):
        """
        :param root:
        :param parent:
        :return:
        """
        name = 'Message %s/%s:%s%%2F%s%%2F%s/%s' % (self.vendor, self.host, self.port, self.name_type, self.name,
                                                    self.role)
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        name = 'GENERAL/Message %s/%s:%s%%2F%s%%2F%s/%s%%2FByte' % (self.vendor, self.host, self.port, self.name_type,
                                                                    self.name, self.role)
        yield TimeMetric(name=name, scope=root.path, duration=self.byte, exclusive=self.exclusive)

        name = "GENERAL/Message %s/%s:%s/%s" % (self.vendor, self.host, self.port, self.role)
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        # 消费者采集他的等待时长
        if self.role == 'Consume':
            name = 'GENERAL/Message %s/%s:%s%%2F%s%%2F%s/%s%%2FWait' % (self.vendor, self.host, self.port,
                                                                        self.name_type, self.name, self.role)
            yield TimeMetric(name=name, scope=root.path, duration=root.trace_interval, exclusive=self.exclusive)

        name = 'GENERAL/Message %s/NULL/All' % self.vendor
        yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        if root.type == 'WebAction':
            name = "GENERAL/Message %s/NULL/AllWeb" % self.vendor
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)
        else:
            name = "GENERAL/Message %s/NULL/AllBackgound" % self.vendor
            yield TimeMetric(name=name, scope=root.path, duration=self.duration, exclusive=self.exclusive)

        for child in self.children:
            for metric in child.time_metrics(root, self):
                yield metric

    def trace_node(self, root):
        """作为调用方时，txData无法回传，此时需要将externalId往服务端传递
        :param root:
        :return:
        """
        start_time = node_start_time(root, self)
        end_time = node_end_time(root, self)
        params = {}
        children = []

        # 当作为调用者时，该数据才会从上报，否则该id会跟着跨应用数据上报服务器
        if self.external_id:
            params['externalId'] = self.external_id

        if root.trace_id:
            params['txId'] = root.trace_id

        root.trace_node_count += 1
        for child in self.children:
            if root.trace_node_count > root.trace_node_limit:
                break

            children.append(child.trace_node(root))

        # 由于MQ产生的跨应用数据无法回传，txdata需要在被调方处理跨应用数据
        name = 'Message %s/%s:%s%%2F%s%%2F%s/%s' % (self.vendor, self.host, self.port, self.name_type, self.name,
                                                    self.role)

        return [start_time, end_time, name, '', 1, self.wrapped_info[0], self.wrapped_info[1], params, children]




# -*- coding: utf-8 -*-

""" define detector for detect tornado httpclient module
"""

import logging
from tingyun.armoury.ammunition.tornado_4.wrappers import wrap_function_wrapper
from tingyun.armoury.ammunition.external_tracker import ExternalTrace
from tingyun.armoury.ammunition.tornado_4.utils import obtain_current_tracker

console = logging.getLogger(__name__)


def parse_request_url(*args, **kwargs):
    """
    """
    from tornado.httpclient import HTTPRequest

    def _parse_request(request, *args, **kwargs):
        return request

    request = _parse_request(*args, **kwargs)
    if isinstance(request, HTTPRequest):
        url = request.url
    else:
        url = request

    return url


def trace_tornado_http_request(wrapped, instance, args, kwargs):
    """
    """
    tracker = obtain_current_tracker()
    if not tracker:
        return wrapped(*args, **kwargs)

    _url = parse_request_url(*args, **kwargs)

    with ExternalTrace(tracker, "httpclient", _url):
        return wrapped(*args, **kwargs)


def detect_http_client(module):
    """
    :param module:
    :return:
    """
    wrap_function_wrapper(module, 'HTTPClient.fetch', trace_tornado_http_request)
    wrap_function_wrapper(module, 'AsyncHTTPClient.fetch', trace_tornado_http_request)




"""used to define console output destination
"""

import sys
import logging
import logging.handlers
from tingyun.packages.wrapt.decorators import synchronized


class _NullHandler(logging.Handler):
    def emit(self, record):
        pass

console = logging.getLogger('tingyun')
console.addHandler(_NullHandler())
_LOG_FORMAT = '%(asctime)s (%(process)d/%(threadName)s) %(name)s %(lineno)s %(levelname)s - %(message)s'
_initialized = False


@synchronized
def initialize_logging(log_file, log_level):
    global _initialized
    if _initialized:
        console.warning("The logging %s was initialized ever.", log_file)
        return 1

    try:
        is_file_log = False
        handler = logging.StreamHandler(sys.stdout)
        if log_file == 'stdout':
            console.info('Initializing Python agent stdout logging.')
        elif log_file == 'stderr':
            handler = logging.StreamHandler(sys.stderr)
            console.info('Initializing Python agent stderr logging.')
        elif log_file:
            try:
                handler = logging.FileHandler(log_file)
                is_file_log = True
                console.info("init log file %s", log_file)
            except Exception as _:
                handler = logging.StreamHandler(sys.stderr)
                console.exception('Create log file %s failed. the tingyun log will be output to stderr.' % log_file)

        handler.setFormatter(logging.Formatter(_LOG_FORMAT))
        console.addHandler(handler)
        console.setLevel(log_level)

        if is_file_log:
            console.propagate = False

        _initialized = True
    except Exception as err:
        console.error("Errors occurred when init agent log. %s", err)

    return 0



# -*- coding: utf-8 -*-

"""the module wrapt main entrance

"""

import logging

from tingyun.armoury.ammunition.mongo_tracker import wrap_mongo_trace
from tingyun.armoury.ammunition.function_tracker import wrap_function_trace

console = logging.getLogger(__name__)
_methods = ['create_index', 'distinct', 'drop', 'drop_index', 'drop_indexes', 'ensure_index', 'find',
            'find_and_modify', 'find_one', 'group', 'index_information', 'inline_map_reduce', 'insert', 'map_reduce',
            'options', 'reindex', 'remove', 'rename', 'save', 'update', 'insert_one', 'insert', 'find_one_and_replace',
            'find_one_and_delete', 'find_one_and_update', 'aggregate']


def detect_connection(module):
    """
    :param module:
    :return:
    """
    if hasattr(module, "Connection"):
        wrap_function_trace(module, 'Connection.__init__', name='%s:Connection.__init__' % module.__name__)


def detect_mongo_client(module):
    """
    :param module:
    :return:
    """
    if hasattr(module, "MongoClient"):
        wrap_function_trace(module, "MongoClient.__init__", name="%s.MongoClient.__init__" % module.__name__)


def detect_collection(module):
    """
    :param module:
    :return:
    """
    def parse_connect_params(collection, *args, **kwargs):
        """collection有个属性database指向Database, Database有个属性client指向MongoClient，MongoClient中有个属性
        _topology_settings指向， 该类有个属性_seeds存储ip跟地址信息
        :param collection:
        :param args:
        :param kwargs:
        :return: (ip, port, collect)
        """
        host, port, db_name = "Unknown", 0, "Unknown"
        if hasattr(collection, "full_name"):
            db_name = collection.full_name

        try:
            # 如果使用默认host、port参数，该包会使用[('localhost', 27017)]，否则为集合set([(host, port)])
            server = collection.database.client._topology_settings._seeds
            if isinstance(server, set):
                server = list(server)

            host, port = server[0]
        except Exception as err:
            console.debug("Can not get the host and port for MongoDB %s", err)

        return host or "Unknown", port or 0, db_name or "Unknown"

    if not hasattr(module, "Collection"):
        console.info("module(%s) has not Collection object.", module)
        return

    for m in _methods:
        if not hasattr(module.Collection, m):
            console.info("Collection has not %s method", m)
            continue

        wrap_mongo_trace(module, "Collection.%s" % m, schema="", method=m, server=parse_connect_params)



# -*- coding: utf-8 -*-

""" 捕获fetch_impl参数中回调函数的参数, 用于获取相应的http header
"""

import logging
from tingyun.armoury.ammunition.tornado_4.utils import obtain_current_tracker
from tingyun.logistics.basic_wrapper import FunctionWrapper, wrap_function_wrapper

console = logging.getLogger(__name__)


def trace_http_fetch_callback(wrapped, instance, args, kwagrs):
    """不捕获其执行时间，仅用于获取http传回的跨应用数据
    :return:
    """
    tracker = obtain_current_tracker()
    if not tracker:
        console.debug("Do not get tracker from current thread, this's maybe finish before.")

    def parse_reponse(response, *_args, **_kwargs):
        """http client(不论是异步还是同步) fetch 发生调用时, 会调用handle_response(response)作为回调函数
        response 为 HTTPResponse 实例
        :return:
        """
        return response, _args, _kwargs

    response, _args, _kwargs = parse_reponse(*args, **kwagrs)
    if hasattr(response, 'headers'):
        tracker._called_traced_data = eval(response.headers.get("X-Tingyun-Tx-Data", '{}'))

    return wrapped(response, _args, _kwargs)


def trace_fetch_impl(wrapped,  instance, args, kwargs):
    """
    :return:
    """
    def parse_callback_func(request, callback, *args, **kwargs):
        """
        :return:
        """
        return request, callback, args, kwargs

    request, callback_func, _args, _kwargs = parse_callback_func(*args, **kwargs)

    return wrapped(request, FunctionWrapper(callback_func, trace_http_fetch_callback), *_args, **_kwargs)


def detect_simple_httpclient(module):
    """
    :return:
    """
    # wrap_function_wrapper(module.Flask, 'fetch_impl', trace_fetch_impl)




"""This module define some user api for manually use the tingyun python agent.

    We will do the best to providing more convenient usage and available description about api, more detail will be as
a extension in user usage documentation
"""

import tingyun.embattle
import tingyun.armoury.trigger.wsgi_entrance
import tingyun.armoury.ammunition.function_tracker


"""
   Description: Init the agent for detecting all of the python plugins that can be detected, more detail listed in
                agent doc `framework support` chapter.)

   Warning: The init action must be called before all of your python code. Or the agent maybe can not work well.

   Usage:

       from tingyun.api import init_tingyun_agent

       init_tingyun_agent(config_file=None)

   Param:
       config_file: The absolutely path of tingyun agent config file.  if the option is not provided, you should set
                    the environment variable `TING_YUN_CONFIG_FILE`, or the agent will not work.
"""
init_tingyun_agent = tingyun.embattle.initialize


"""
    Description: A wrapper only used to wsgi application entrance.

    Usage:

        # init the tingyun agent first
        from tingyun.api import init_tingyun_agent
        from tingyun.api import wsgi_app_decorator

        init_tingyun_agent(config_file='/tmp/tingyun.ini')

        @wsgi_app_decorator(framework='xx', version='xx')
        def application(environ, start_response):
            status = '200 OK' # HTTP Status
            headers = [('Content-type', 'text/plain')] # HTTP Headers
            start_response(status, headers)

            # The returned object is going to be printed
            return ["Hello World"]

    Params:
        framework: this param indicate witch framework your application used. decorator has default value-`xx` for
                   this parameter.
        version: the version of the framework. decorator has default value-`xx` for this parameter.
"""
wsgi_app_decorator = tingyun.armoury.trigger.wsgi_entrance.wsgi_application_decorator


"""
    Description: A wrapper used to collect function metric performance data. this function should be used after function
                 init_tingyun_agent() and wsgi_app_decorator()

    Usage:
        from tingyun.api import init_tingyun_agent
        from tingyun.api import wsgi_app_decorator
        from tingyun.api import function_trace_decorator

        init_tingyun_agent(config_file='/tmp/tingyun.ini')

        @wsgi_app_decorator(framework='xx', version='xx')
        def application(environ, start_response):
            status = '200 OK' # HTTP Status
            headers = [('Content-type', 'text/plain')] # HTTP Headers
            start_response(status, headers)

            return ["Hello World"]

        import hashlib

        @function_trace_decorator(name='md5sum', group='calculator')
        def md5sum(filename):
            fd = open(filename,"r")
            fcont = fd.r
            fd.close()
            fmd5 = hashlib.md5(fcont)
            return fmd5

    Params: both parameter can be `None`, we will set default value for it.
        name: the name of the function or some meaningful to you.
        group: the group of the function belong to. or some meaningful to you
"""
function_trace_decorator = tingyun.armoury.ammunition.function_tracker.function_trace_decorator



""" define this module for do the thread profile analyze

"""
import threading
import logging
import time

try:
    from sys import intern
except ImportError:
    pass

from tingyun.config.settings import global_settings
from tingyun.battlefield.knapsack import knapsack

console = logging.getLogger(__name__)
RUNNABLE_STATUS_FUNC = ['wait', 'sleep']


class ProfileTraceNode(object):
    """
    """
    def __init__(self, category="", class_name="", method_name="", line_num=0, is_runnable=None):
        """
        """
        self.runnable = 0
        self.non_runnable = 0
        self.class_name = class_name
        self.method_name = method_name
        self.line_num = line_num
        self.category = category
        self.depth = 0
        self.retrieve_depth = 0

        self.children = {}  # {func_name: node}
        self.is_runnable = is_runnable  # {"is_runnable": False} just for profile data node loop used to mark.

    @property
    def trace_segment(self):
        """
        """
        return [self.class_name, self.method_name, self.line_num]

    def has_child(self):
        """
        """
        if len(self.children) > 0:
            return True

        return False

    def format_data(self):
        """ just return the basic data, not include child node
        """
        return [self.trace_segment, self.runnable, self.non_runnable]

    def append_child(self, key, child):
        """
        """
        self.children[key] = child

    def get_child_key(self):
        """ a new trace node should has one child at most.
        """
        if not self.has_child():
            return 'unknown'

        return self.children.keys()[0]


class Profile(object):
    """
    """

    def __init__(self, profile_id):
        """
        """
        self.profile_id = profile_id
        self.start_time = int(time.time() * 1000)
        self.end_time = 0
        self.final_data = {"Web": {"cpuTime": 0, "threadTraces": []}, "Background": {"cpuTime": 0, "threadTraces": []},
                           "Agent": {"cpuTime": 0, "threadTraces": []}, "Other": {"cpuTime": 0, "threadTraces": []}}
        self.sample_count = 0
        self.thread_count = 0
        self.total_runnable_count = 0
        self.stacktraces = {}
        self.max_retrieve_depth = global_settings().max_profile_depth  # avoid reaching python max recursion depth.
        self.retrieve_depth = 0

    def profile_data(self):
        """Note, we do not update the root node basic data.(runnable,non-runnable)
        The root node dose not include the thread data.

        """
        for category in self.final_data:

            if (category not in self.stacktraces) or (not self.stacktraces[category].has_child()):
                console.info("Get empty data with category: %s", category)
                self.final_data[category]['threadTraces'] = []
                continue

            # retrieve_profile_node() will lost the first node of the profile. so rebuild default one
            # if node in self.stacktraces
            root = [['-', 'PythonProfileEntryFunc', 1], 1, 1, []]
            self.retrieve_profile_node(self.stacktraces[category], root)
            self.final_data[category]['threadTraces'] = root[3]

        payload = {"profileId": self.profile_id, "beginTime": self.start_time, "endTime": self.end_time,
                   "sampleCount": self.sample_count, "threads": self.thread_count,
                   "stacktraces": str(self.final_data), "runnableThreads": self.total_runnable_count}
        return payload

    def retrieve_profile_node(self, profile_node, parent):
        """
        :param profile_node:
        :return:
        """
        self.retrieve_depth += 1

        if not profile_node.has_child() or profile_node.retrieve_depth > self.max_retrieve_depth:
            parent[3].append(profile_node.format_data() + [[]])
            return

        for key in profile_node.children:
            node = profile_node.children[key]

            data_node = node.format_data() + [[]]
            parent[3].append(data_node)

            self.retrieve_profile_node(node, data_node)

    def update_thread_trace(self, category, new_node):
        """
        :param category:
        :param new_node: new stack trace node
        :return:
        """
        if category != new_node.category:
            return

        # very important. node with head node without data maybe cause data lost.
        if not new_node.has_child:
            console.info("New profile node is empty, ignore it now.")
            return

        key = category
        if key not in self.stacktraces:
            self.update_trace_node_runnable_status(new_node)
            self.stacktraces[key] = new_node
            # console.info('new tree with category: %s, data: %s', category, new_node)
            return

        # it's dict
        root_tree = self.stacktraces[key]

        while True:
            # only one child exist if new node has child. this will drop the first node.
            if not new_node.has_child():
                break

            # note: the top level node of the new node is a empty node.
            # It's new branch of the root node. so break out
            # root_tree can not use the append_child method, because the current node is lost just get the child(dict)
            child_key = new_node.get_child_key()
            if child_key not in root_tree.children:
                self.update_trace_node_runnable_status(new_node.children[child_key])
                root_tree.children[child_key] = new_node.children[child_key]
                break

            # update the node attribution
            if new_node.is_runnable["is_runnable"]:
                root_tree.children[child_key].runnable += 1
            else:
                root_tree.children[child_key].non_runnable += 1

            # deal the child node.
            root_tree = root_tree.children[child_key]
            new_node = new_node.children[child_key]

    def update_trace_node_runnable_status(self, trace_node):
        """ the trace node must be single tree switch. it is used in new thread trace node
        :param trace_node:
        :return:
        """
        runnable = trace_node.is_runnable["is_runnable"]
        root = trace_node

        while root:
            if runnable:
                root.runnable = 1
            else:
                root.non_runnable = 1

            if root.has_child():
                root = root.children.values()[0]
            else:
                break


class ProfileManager(object):
    """Note: Now we are not support multi application. only use the profile id as verification.
    """
    lock = threading.Lock()
    instance = None

    def __init__(self):
        """we provide local setting switch, just used when exception occurred in user environment.
        enable parameter will be removed in the future.
        :return:
        """
        self.enable = global_settings().enable_profile
        self.current_profile = None

        self._profile_shutdown = threading.Event()
        self.profile_thread_running = False  # not running
        self.profile_thread = None
        self.cmd_info = {}
        self.finished_profiles = []
        self.cmd_stop_status = False
        self.generated_data = None

    @staticmethod
    def singleton():
        with ProfileManager.lock:
            if ProfileManager.instance is None:
                ProfileManager.instance = ProfileManager()

        return ProfileManager.instance

    def start_profile(self, cid, app_name, profile_id, duration, interval):
        """
        :param app_name:
        :param profile_id:
        :param duration: sample the profile duration, 'seconds' format
        :param interval: mile seconds format
        :return: True/False
        """
        self.cmd_info['cid'] = cid
        self.cmd_info['profile_id'] = profile_id
        self.cmd_info['duration'] = duration
        self.cmd_info['interval'] = interval / 1000.0
        self.cmd_info['app_name'] = app_name
        self.cmd_info['end_time'] = time.time() + duration

        if not self.enable:
            console.info("profile disabled in local settings.")
            return False

        if self.current_profile:
            console.warning("There's maybe running a profile, this is a unnormal status, if this continues,"
                            "please report us for further investigation.")
            return False

        with self.lock:
            if self.profile_thread_running:
                console.warning("There are running a thread profile, this is a unnormal status, if this continues,"
                                "please report us for further investigation.")
                return False

            self.current_profile = Profile(profile_id)

            self.profile_thread = threading.Thread(target=self._do_profile, name="TingYun-profile-thread")
            self.profile_thread.setDaemon(True)
            self.profile_thread.start()

            self.profile_thread_running = True  # running.

        return True

    def stop_profile(self):
        """
        :return:
        """
        with self.lock:
            if self.current_profile:
                self.current_profile.end_time = int(time.time() * 1000)
                self.finished_profiles.append(self.current_profile)

            self.current_profile = None
            self.profile_thread_running = False

    def _do_profile(self):
        """
        :return:
        """
        console.info("Staring thread for doing profile.")
        while True:
            if not self.current_profile:
                console.info("Stopping profile now.")
                return

            self.current_profile.sample_count += 1
            for thread_id, category, stack_node in self.trace_thread_stack():
                if not stack_node.has_child():
                    console.debug("Do not get the thread stacktrace.")
                    continue

                self.current_profile.update_thread_trace(category, stack_node)

            self.update_profile_status()

            interval = 0.1 if not self.cmd_info["interval"] else self.cmd_info["interval"]
            self._profile_shutdown.wait(interval)

    def update_profile_status(self):
        """
        :return:
        """
        if self.current_profile:
            if time.time() > self.cmd_info['end_time']:
                self.stop_profile()
                console.info("Finish the profile thread.")

    def trace_thread_stack(self):
        """
        :return:
        """
        for tracker, thread_id, category, frame in knapsack().get_active_threads():
            # top level node for profile
            is_runnable = {"is_runnable": True}
            top_node = ProfileTraceNode(category, "-", 'PythonProfileEntryFunc', 0, is_runnable)
            father = top_node
            self.current_profile.thread_count += 1
            while frame:
                code = frame.f_code
                child = ProfileTraceNode(category, intern(code.co_filename), intern(code.co_name), frame.f_lineno,
                                         is_runnable)
                father.append_child(intern(code.co_name), child)
                father = child
                top_node.depth += 1
                frame = frame.f_back

                # mark it is runnable thread or not
                if intern(intern(code.co_name)) in RUNNABLE_STATUS_FUNC:
                    is_runnable['is_runnable'] = False
                    break

            if is_runnable['is_runnable']:
                self.current_profile.total_runnable_count += 1

            yield thread_id, category, top_node

    def generate_profile_data(self):
        """
        :return:
        """
        if not self.enable:
            profile = Profile(self.cmd_info['profile_id'])
            self.finished_profiles.append(profile)

        # dose not finished
        if self.current_profile:
            return None

        # server stop command
        if self.cmd_stop_status:
            return None

        if 0 == len(self.finished_profiles):
            console.warning("We got empty profile data, this maybe some issues of the python code, if this continues, "
                            "please report to us ofr further investigation. Thank U.")
            console.warning("Current command information is %s", self.cmd_info)
            return None

        try:
            final_data = {"id": self.cmd_info['cid'], "result": self.finished_profiles[0].profile_data()}
        except Exception as err:
            console.warning("Errors, when generate profile data. %s, %s", self.cmd_info, err)
            self.finished_profiles = []
            return None

        self.finished_profiles = []

        return final_data

    def shutdown(self, stop_cmd=False):
        """
        :return:
        """
        console.info("Force shutdown the profile, because receive the shutdown signals.")
        self.stop_profile()
        self.finished_profiles = []
        if stop_cmd:
            self.cmd_stop_status = True

        return True


def get_profile_manger():
    """only singleton mode used.
    :return:
    """
    return ProfileManager.singleton()



# -*- coding: utf-8 -*-

"""The detect module follow the protocol: https://www.python.org/dev/peps/pep-0249/

"""

import re
import logging

from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.database_tracker import DatabaseTracker
from tingyun.armoury.ammunition.function_tracker import FunctionTracker, wrap_function_trace

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse


# module_name: standard db name for data exchange protocol.
# 数据库类型协议，参考文档
DB_TYPE_MYSQL = "MySQL"
DB_TYPE_ORACLE = "Oracle"
DB_TYPE_POSTGRELSQL = "PostgreSQL"
DB_TYPE_SQLSERVER = "SQLServer"
db_name_module = {
    "MySQLdb": DB_TYPE_MYSQL, "pymysql": DB_TYPE_MYSQL, "oursql": "MySQL",
    "cx_Oracle": DB_TYPE_ORACLE,
    "psycopg2": DB_TYPE_POSTGRELSQL, "psycopg2ct": DB_TYPE_POSTGRELSQL, "psycopg2cffi": DB_TYPE_POSTGRELSQL,
    "pyodbc": DB_TYPE_SQLSERVER
    }
console = logging.getLogger(__name__)


def detect(module):
    """ more interface description about dbapi2 in: https://www.python.org/dev/peps/pep-0249/
    :param module:
    :return:
    """
    dbtype = db_name_module.get(getattr(module, "__name__", "DBAPI2"))

    def parse_connect_params(connect_params):
        """解析连接数据库的host、port信息，支持mysql-python、pymysql、oursql、cx_Oracle插件
            根据数据库分类来分类获取port信息
            mysql-python:
                connect(host, user, pwd, port, db)
            pymysql
                connect(host=None, port=0, database=None)
            oursql http://pythonhosted.org/oursql：
                oursql.connect(host='127.0.0.1', user='habnabit', passwd='foobar',
                db='example', port=3307)
        :param connect_params:
        :return:
        """
        host, port, db_name = 'Unknown', 'Unknown', 'Unknown'
        if not connect_params or len(connect_params) != 2:
            return

        _args = connect_params[0] or ()
        _kwargs = connect_params[1] or {}

        if dbtype == DB_TYPE_MYSQL:
            # 遵循DBAPI2协议前5个参数都是host=None, user=None, password="", database=None, port=0
            host, port = _kwargs.get('host'), _kwargs.get('port')
            db_name = _kwargs.get('database') or _kwargs.get('db')

            if not host:
                host = _args[0] if len(_args) >= 1 else "Unknown"

            if not port:
                port = _args[4] if len(_args) >= 5 else 3306
            else:
                port = 3306

            if not db_name:
                db_name = _args[3] if len(_args) >= 4 else "Unknown"

        elif dbtype == DB_TYPE_ORACLE:
            # cx_Oracle.connect('hr', 'hrpwd', 'localhost:1521/XE') 或者
            # '(DESCRIPTION=(ADDRESS_LIST=(ADDRESS=(PROTOCOL=TCP)(HOST=oracle.db.com)(PORT=1521)))\
            # (CONNECT_DATA=(SID=nbsdb)))'
            if 3 == len(_args):
                con_info = re.findall(r'^(.+?):(.+?)/(.+?)$', str(_args[2]), re.IGNORECASE)
                if not con_info:
                    con_info = re.findall(r'^.*HOST=(.+?)\).*PORT=(.+?)\).*SID=(.+?)\).?', str(_args[2]), re.IGNORECASE)

                if con_info and 3 == len(con_info[0]):
                    host, port, db_name = con_info[0]

            # 'hr/hrpwd@localhost:1521/XE'
            if 1 == len(_args):
                con_str = str(_args[0])
                _host = con_str[con_str.find("@") + 1: con_str.find(":")]
                _port = con_str[con_str.find(":") + 1: con_str.rfind("/")]
                _db_name = con_str[con_str.rfind("/") + 1:]

                if _host and _port and _db_name:
                    host, port, db_name = _host, _port, _db_name

        elif dbtype == DB_TYPE_POSTGRELSQL:
            # 链接方式doc: https://www.postgresql.org/docs/current/static/libpq-connect.html#LIBPQ-CONNSTRING
            # postgresql://localhost/mydb
            # postgresql://localhost:5432/mydb
            # dsn参数: host=localhost port=5432 dbname=mydb connect_timeout=10
            # 关键字参数
            # todo: 不支持以下链接方式：
            # 不完整的URI连接， postgresql:///mydb?host=localhost&port=5433
            # IPV4, postgresql://[2001:db8::1234]/database
            # 环境变量方式： https://www.postgresql.org/docs/current/static/libpq-envars.html
            _host, _port, _db_name = None, None, None

            if not _kwargs or len(_kwargs) == 1:
                con_str = _args[0] if 1 == len(_args) and not _kwargs else _kwargs.get('dsn', "")

                # host=localhost port=5432 dbname=mydb connect_timeout=10
                if "//" not in con_str:
                    for kv in con_str.split():
                        k, v = kv.split('=')
                        if k.lower() not in ["host", "port", "dbname"]:
                            continue

                        if 'host' == k.lower():
                            _host = v
                        elif 'port' == k.lower():
                            _port = v
                        elif 'dbname' == k.lower():
                            _db_name = v

                else:
                    # postgresql://[user[:password]@][netloc][:port][/dbname][?param1=value1&...]
                    con_str = _args[0] if 1 == len(_args) and not _kwargs else _kwargs.get('dsn', "")
                    sections = urlparse.urlparse(con_str)

                    if sections.netloc:
                        if "@" in sections.netloc:
                            host_port = sections.netloc.split("@")[1]
                        else:
                            host_port = sections.netloc

                        parts = host_port.split(":")
                        if len(parts) == 2:
                            _host, _port = parts[0], parts[1]
                        elif len(parts) == 1:
                            _host, _port = parts[0], 5432

                        _db_names = sections.path.replace('/', '', 1).split('?')
                        _db_name = _db_names[0] if len(_db_names) >= 1 else sections.path
            else:
                _host, _port = _kwargs.get("host"), _kwargs.get("port") or 5432
                _db_name = _kwargs.get("database") or _kwargs.get("dbname")

            host, port, db_name = _host or 'Unknown', _port or 5432, _db_name or 'Unknown'

        elif dbtype == DB_TYPE_SQLSERVER:
            # pyodbc 包仅支持sqlServer数据的抓取，其他类型无法识别，也几乎没人会用这么复杂的包连接别的数据库
            if len(_args) >= 1 and 'DSN' not in str(_args[0]).upper():
                # 'DSN=test;PWD=password',如果用户使用了DSN，无法抓取该信息
                # ('DRIVER={FreeTDS};SERVER=sqlserver.db.com; PORT=1433;DATABASE=python_test;
                # UID=leng;PWD=Windows2008',)
                # 仅支持表中的一种port，host写法,不区分大小写
                _host = re.findall(r"SERVER=(.+?);", str(_args[0]), re.IGNORECASE)
                _port = re.findall(r"port=(.+?);", str(_args[0]), re.IGNORECASE)
                _db_name = re.findall(r"DATABASE=(.+?);", str(_args[0]), re.IGNORECASE)

                _host, _port = _host[0] if _host else None, _port[0] if _port else None
                _db_name = _db_name[0] if _db_name else None

                # 会出现'DRIVER={FreeTDS};SERVER=sqlserver.db.com,1433;DATABASE=python_test;UID=leng;PWD=Windows2008'
                if _host and _db_name and not _port:
                    parts = str(_host).split(",")
                    if 2 == len(parts):
                        _host = parts[0].strip()
                        _port = parts[1].strip()

                host, port, db_name = _host or 'Unknown', _port or 'Unknown', _db_name or 'Unknown'
        else:
            logging.debug("Not support database type for capture ip & port.")

        return host, port, db_name

    class TingYunCursor(object):

        def __init__(self, cursor, cursor_params=None, connect_params=None):
            """https://docs.python.org/2.7/reference/datamodel.html?highlight=object.__setattr__#object.__setattr__
            :param cursor:
            :param cursor_params:
            :param connect_params:
            :return:
            """
            object.__setattr__(self, 'cursor', cursor)
            object.__setattr__(self, 'cursor_params', cursor_params)
            object.__setattr__(self, 'connect_params', connect_params)

            host, port, db_name = parse_connect_params(connect_params)

            object.__setattr__(self, 'host', host)
            object.__setattr__(self, 'port', port)
            object.__setattr__(self, 'db_name', db_name)

        def __setattr__(self, name, value):
            setattr(self.cursor, name, value)

        def __getattr__(self, name):
            return getattr(self.cursor, name)

        def __iter__(self):
            return iter(self.cursor)

        def fetchone(self, *args, **kwargs):
            """we do not capture the metric of execute result. this is small time used
            :args:
            :kwargs:
            :return:
            """
            return self.cursor.fetchone(*args, **kwargs)

        def fetchmany(self, *args, **kwargs):
            """we do not capture the metric of execute result. this is small time used
            :args:
            :kwargs:
            :return:
            """
            return self.cursor.fetchmany(*args, **kwargs)

        def fetchall(self, *args, **kwargs):
            """this operation maybe spend more time. this is small time used
            and the sql was executed. we can not take it
            :args:
            :kwargs:
            :return:
            """
            return self.cursor.fetchall(*args, **kwargs)

        def execute(self, sql, *args, **kwargs):
            """
            :param sql:
            :param args:
            :param kwargs:
            :return:
            """
            tracker = current_tracker()
            if not tracker:
                return self.cursor.execute(sql, *args, **kwargs)

            with DatabaseTracker(tracker, sql, dbtype, module, self.connect_params, self.cursor_params,
                                 (args, kwargs), host=self.host, port=self.port, db_name=self.db_name):
                return self.cursor.execute(sql, *args, **kwargs)

        def executemany(self, sql, *args, **kwargs):
            """
            :param sql:
            :param args:
            :param kwargs:
            :return:
            """
            tracker = current_tracker()
            if not tracker:
                return self.cursor.executemany(sql, *args, **kwargs)

            with DatabaseTracker(tracker, sql, dbtype, module, host=self.host, port=self.port, db_name=self.db_name):
                return self.cursor.executemany(sql, *args, **kwargs)

        def callproc(self, procname, *args, **kwargs):
            """
            :param procname:
            :param args:
            :param kwargs:
            :return:
            """
            tracker = current_tracker()
            if not tracker:
                return self.cursor.callproc(procname, *args, **kwargs)

            with DatabaseTracker(tracker, 'CALL %s' % procname, dbtype, module, host=self.host, port=self.port,
                                 db_name=self.db_name):
                return self.cursor.callproc(procname, *args, **kwargs)

    class TingYunConnection(object):

        def __init__(self, connection, params=None):
            """
            :param connection:
            :param params:
            :return:
            """
            object.__setattr__(self, 'connection', connection)
            object.__setattr__(self, 'params', params)
            host, port, db_name = parse_connect_params(params)

            object.__setattr__(self, 'host', host)
            object.__setattr__(self, 'port', port)
            object.__setattr__(self, 'db_name', db_name)

        def __setattr__(self, name, value):
            setattr(self.connection, name, value)

        def __getattr__(self, name):
            return getattr(self.connection, name)

        def close(self, *args, **kwargs):
            """
            :param args:
            :param kwargs:
            :return:
            """
            return self.connection.close(*args, **kwargs)

        def cursor(self, *args, **kwargs):
            """
            :param args:
            :param kwargs:
            :return:
            """
            return TingYunCursor(self.connection.cursor(*args, **kwargs), (args, kwargs), self.params)

        def commit(self):
            """
            :return:
            """
            tracker = current_tracker()
            if not tracker:
                return self.connection.commit()

            with DatabaseTracker(tracker, 'COMMIT', dbtype, module, host=self.host, port=self.port,
                                 db_name=self.db_name):
                return self.connection.commit()

        def rollback(self):
            """
            :return:
            """
            tracker = current_tracker()
            if not tracker:
                return self.connection.rollback()

            with DatabaseTracker(tracker, 'ROLLBACK', dbtype, module, host=self.host, port=self.port,
                                 db_name=self.db_name):
                return self.connection.rollback()

    class DatabaseWrapper(object):

        def __init__(self, connect):
            self.connect = connect
            self.connect_instance = None

        def __call__(self, *args, **kwargs):
            """when module call the connect method. the wrapper will callback __call__ instead.
            :param args:
            :param kwargs:
            :return:
            """
            self.connect_instance = self.connect(*args, **kwargs)

            return TingYunConnection(self.connect_instance, (args, kwargs))

    # Check if module is already wrapped
    if hasattr(module, '_self_dbapi2_wrapped'):
        return

    wrap_function_trace(module, 'connect', name='%s.connect' % dbtype, group='Database')
    module.connect = DatabaseWrapper(module.connect)
    module._self_dbapi2_wrapped = True



"""This module provides functions to collect information about the operating system, Python and hosting environment.

"""

import os
import sys
import platform
from tingyun.armoury.sampler.system_info import cpu_count, memory_total, cpu_info

try:
    import pkg_resources
except ImportError:
    pass


def env_config():
    """
    :return: env, a dict of environment
    """
    env = dict()

    # system information
    env["OS Arch"] = platform.machine()
    env["kernel"] = platform.release()
    env["CPU Cores"] = cpu_count()
    env["Physical Memory"] = "%sMB" % memory_total()
    env["OS Version"] = platform.version()

    # python information
    env["Python Program Name"] = sys.argv[0]
    env["Python Executable"] = sys.executable

    env["Python Home"] = os.environ.get('PYTHONHOME', '')
    env["Python Path"] = os.environ.get('PYTHONPATH', '')

    env["Python Prefix"] = sys.prefix
    env["Python Exec Prefix"] = sys.exec_prefix

    env["Python Version"] = sys.version
    env["Python Platform"] = sys.platform
    env["Python Implementation"] = platform.python_implementation()  # new in python2.6
    env["Python Version"] = platform.python_version()

    env["Python Max Unicode"] = sys.maxunicode
    env["CPU Vendor"], env["CPU Model"], env["CPU mhz"] = cpu_info()

    # Dispatcher information.
    dispatcher = {}
    if not dispatcher and 'mod_wsgi' in sys.modules:
        mod_wsgi = sys.modules['mod_wsgi']
        if hasattr(mod_wsgi, 'process_group'):
            if mod_wsgi.process_group == '':
                dispatcher["Dispatcher"] = 'Apache/mod_wsgi (embedded)'
            else:
                dispatcher["Dispatcher"] = 'Apache/mod_wsgi (daemon)'
        else:
            dispatcher["Dispatcher"] = 'Apache/mod_wsgi'

        if hasattr(mod_wsgi, 'version'):
            dispatcher["Dispatcher Version"] = str(mod_wsgi.version)

    if not dispatcher and 'uwsgi' in sys.modules:
        dispatcher["Dispatcher"] = 'uWSGI'
        uwsgi = sys.modules['uwsgi']

        if hasattr(uwsgi, 'version'):
            dispatcher["Dispatcher Version"] = uwsgi.version

    env.update(dispatcher)

    plugins = []
    tingyun = []

    # Do NOT use six module for the following iteration.
    for name, module in list(sys.modules.items()):
        if name.startswith('tingyun.hooks'):
            tingyun.append("%s." % name)

        elif name.find('.') == -1 and hasattr(module, '__file__'):
            try:
                if 'pkg_resources' in sys.modules:
                    version = pkg_resources.get_distribution(name).version
                    if version:
                        name = '%s(%s).' % (name, version)
            except Exception as _:
                name = '%s.' % name

            plugins.append(name)

    tingyun.extend(plugins)
    env['Plugin List'] = ','.join(tingyun)

    return env


if __name__ == "__main__":
    print "self test: %s" % env_config()



# -*- coding: utf-8 -*-

"""支持RabbitMQ数据采集
"""

import sys
import logging

from tingyun.armoury.mq.utils import retrieve_tracker, mq_common_callback_wrapper

from tingyun.armoury.ammunition.mq_tracker import MQTrace
from tingyun.armoury.ammunition.tracker import current_tracker
from tingyun.armoury.ammunition.external_tracker import process_cross_trace
from tingyun.armoury.ammunition.function_tracker import wrap_function_trace, FunctionTracker
from tingyun.logistics.object_name import callable_name
from tingyun.logistics.basic_wrapper import wrap_object, FunctionWrapper, wrap_function_wrapper

console = logging.getLogger(__name__)


def mq_produce_trace_wrapper(wrapped):
    def dynamic_wrapper(wrapped, instance, args, kwargs):
        tracker = current_tracker()
        vendor = 'RabbitMQ'
        name_type = 'Exchange'
        publish_headers = {}
        role = 'Produce'

        if not tracker:
            return wrapped(*args, **kwargs)

        def parse_server(instance, exchange, routing_key, body, properties=None, *_args, **_kwargs):
            """ 获取publish的一些必要信息，优先获取exchange作为metric名字
            :return:
            """
            global publish_headers
            host, port, byte, name = "Unknown", 0, 0, "Unknown"

            try:
                # blocking  和 其他异步生产者，
                connection = getattr(instance, 'connection', None) or getattr(instance, '_connection', None)
                impl = getattr(connection, '_impl', None) or connection
                host = impl.params.host
                port = impl.params.port

                publish_headers = getattr(properties, "headers", {}) or {}
                ty_headers, _external_id = process_cross_trace(None, msg_type='TingyunID')
                if ty_headers:
                    from pika.spec import BasicProperties

                    if not properties:
                        properties = BasicProperties(headers=ty_headers)
                    else:
                        if not getattr(properties, "headers", None):
                            setattr(properties, "headers", ty_headers)
                        elif isinstance(properties.headers, dict):
                            properties.headers.update(ty_headers)

            except Exception as err:
                console.info("Parse RabbitMQ host & port with error %s", err)
                _external_id = None

            _server = (host, port, sys.getsizeof(body), exchange or routing_key, _external_id)
            return _server, exchange, routing_key, body, properties, _args, _kwargs

        server, exchange, routing_key, body, properties, _args, _kwargs = parse_server(instance, *args, **kwargs)
        host, port, byte, name, external_id = server
        name_type = 'Queue' if not exchange else name_type

        with MQTrace(tracker, vendor, name_type, host, port, byte, name, publish_headers, role,
                     ('BlockingChannel', 'publish'), external_id):
            with FunctionTracker(tracker, callable_name(wrapped)):
                return wrapped(exchange, routing_key, body, properties, *_args, **_kwargs)

    return FunctionWrapper(wrapped, dynamic_wrapper)


def callback_wrapper(wrapped, instance, args, kwargs):
    """
    :return:
    """
    def parse_callback_params(_channel, _method, _properties, _body, *_args, **_kwarg):
        """细节参考pika官方文档http://pika.readthedocs.io/en/0.10.0/examples/blocking_consumer_generator.html
        :return:
        """
        return _channel, _method, _properties, _body, _args, _kwarg

    def parse_server_info(ch):
        """
        :return:
        """
        try:
            connection = getattr(ch, 'connection', None) or getattr(ch, '_connection', None)
            impl = getattr(connection, '_impl', None) or connection
            _host = impl.params.host
            _port = impl.params.port
        except Exception as err:
            _host, _port = "Unknown", 0
            console.debug("Parsing host & port failed when trace consume callback. %s, %s", wrapped, err)

        return _host, _port

    _ch, method, properties, body, _args, _kwarg = parse_callback_params(*args, **kwargs)
    tracker, is_web = retrieve_tracker(getattr(properties, "headers", {}))
    if not tracker:
        return wrapped(*args, **kwargs)

    vendor = 'RabbitMQ'
    name_type = 'Exchange' if method.exchange else "Queue"
    role = 'Consume'

    receive_headers = dict()
    receive_headers["message"] = {}
    receive_headers["message"]["message.byte"] = sys.getsizeof(body)
    receive_headers["message"]["message.queue"] = ''
    receive_headers["message"]["message.exchange"] = method.exchange
    receive_headers["message"]["message.routingkey"] = method.routing_key

    request_params = getattr(properties, "headers", {})
    receive_headers.update(request_params or {})
    host, port = parse_server_info(_ch)

    exceptions = None
    with MQTrace(tracker, vendor, name_type, host, port, sys.getsizeof(body), method.exchange or method.routing_key,
                 receive_headers, role, ('', wrapped.__name__), None):
        with FunctionTracker(tracker, callable_name(wrapped)):
            try:
                result = wrapped(*args, **kwargs)
                tracker.deal_response("200 ok", {})
            except Exception as err:  # Catch all
                tracker.record_exception(*sys.exc_info())
                tracker.deal_response("500 error", {})
                exceptions = err

    # 不在web应用内
    if not is_web:
        tracker._parse_request_params(receive_headers)
        tracker.set_tracker_name("%s%%2F%s" % (name_type, method.exchange or method.routing_key), priority=2)
        tracker.finish_work(*sys.exc_info())
    if exceptions:
        raise exceptions

    return result


def mq_consume_wrapper(wrapped, instance, args, kwargs):
    """
    :return:
    """
    def parse_callback(cb, *_args, **_kwargs):
        """
        :return:
        """
        return cb, _args, _kwargs

    if 'consumer_callback' in kwargs:
        callback_class_name = getattr(getattr(kwargs['consumer_callback'], 'im_class', None), '__name__', None)
        if getattr(kwargs['consumer_callback'], '_self_is_wrapped', None) or callback_class_name == 'BlockingChannel':
            console.debug("Consumer was wrapped before.")
            return wrapped(*args, **kwargs)

        kwargs['consumer_callback'] = FunctionWrapper(kwargs['consumer_callback'], callback_wrapper)
        return wrapped(*args, **kwargs)
    else:
        cb, _args, _kwargs = parse_callback(*args, **kwargs)
        callback_class_name = getattr(getattr(cb, 'im_class', None), '__name__', None)
        if getattr(cb, '_self_is_wrapped', None) or callback_class_name == 'BlockingChannel':
            console.debug("Consumer was wrapped before.")
            return wrapped(cb, *_args, **_kwargs)

        wrapped_cb = FunctionWrapper(cb, callback_wrapper)
        wrapped_cb._self_is_wrapped = True

        return wrapped(wrapped_cb, *_args, **_kwargs)


def detect_block_channel(module):
    """
    :return:
    """
    # wrap_object(module, "BlockingChannel.publish", mq_produce_trace_wrapper)
    wrap_function_wrapper(module.BlockingChannel, "basic_consume", mq_consume_wrapper)


def detect_channel(module):
    """
    :return:
    """
    # 基于selection connection libevent connect tornado connection 的生产者目前使用场景大部分为非web过程内
    # 需要单独处理才能支持，咱不考虑
    wrap_object(module, "Channel.basic_publish", mq_produce_trace_wrapper)
    wrap_function_wrapper(module.Channel, "basic_consume", mq_consume_wrapper)

    # 下面的方法都是采用即时读写的方式跟mq通信，故可以直接采集之性能
    wrap_function_trace(module.Channel, "basic_get", name='basic_get')
    wrap_function_trace(module.Channel, "basic_nack", name='basic_nack')
    wrap_function_trace(module.Channel, "exchange_bind", name='exchange_bind')
    wrap_function_trace(module.Channel, "exchange_declare", name='exchange_declare')
    wrap_function_trace(module.Channel, "exchange_delete", name='exchange_delete')
    wrap_function_trace(module.Channel, "queue_bind", name='queue_bind')
    wrap_function_trace(module.Channel, "queue_declare", name='queue_declare')
    wrap_function_trace(module.Channel, "queue_delete", name='queue_delete')

    # 其他常规用户可能执行的回调
    # 为了减少性能开销，不捕获基于CallbackManager.add所有的回调函数
    wrap_function_wrapper(module, "Channel.add_on_cancel_callback", mq_common_callback_wrapper)
    wrap_function_wrapper(module, "Channel.add_on_close_callback", mq_common_callback_wrapper)
    wrap_function_wrapper(module, "Channel.add_on_return_callback", mq_common_callback_wrapper)


def detect_connection(module):
    """"""
    # 其他常规用户可能执行的回调
    # blockingconnection中的add_on_connection_blocked_callback&
    # add_on_connection_unblocked_callback都是调用connection中同样的方法
    # 为了减少性能开销，不捕获基于CallbackManager.add所有的回调函数
    wrap_function_wrapper(module, "Connection.add_on_close_callback", mq_common_callback_wrapper)
    wrap_function_wrapper(module, "Connection.add_on_connection_blocked_callback", mq_common_callback_wrapper)
    wrap_function_wrapper(module, "Connection.add_on_connection_unblocked_callback", mq_common_callback_wrapper)
    wrap_function_wrapper(module, "Connection.add_on_open_callback", mq_common_callback_wrapper)
    wrap_function_wrapper(module, "Connection.add_on_open_error_callback", mq_common_callback_wrapper)
